# -*- coding: utf-8 -*-
# Owlready2
# Copyright (C) 2013-2019 Jean-Baptiste LAMY
# LIMICS (Laboratoire d'informatique médicale et d'ingénierie des connaissances en santé), UMR_S 1142
# University Paris 13, Sorbonne paris-Cité, Bobigny, France

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import owlready2, owlready2.editor
from owlready2 import *
from owlready2.instance_editor import *

import editobj3, editobj3.undoredo as undoredo, editobj3.introsp as introsp, editobj3.field as field, editobj3.editor as editor, editobj3.editor_qt as editor_qt

import PyQt5.QtCore    as qtcore
import PyQt5.QtWidgets as qtwidgets
import PyQt5.QtGui     as qtgui


class QtOntologyInstanceEditor(OntologyInstanceEditor, editor_qt.QtEditorTabbedDialog):    
  def check_save(self):
    if self.undo_stack.undoables != self.last_undoables:
      box = qtwidgets.QMessageBox()
      box.setText(editobj3.TRANSLATOR(u"Save modifications before closing?"))
      box.setStandardButtons(qtwidgets.QMessageBox.Save | qtwidgets.QMessageBox.Discard | qtwidgets.QMessageBox.Cancel)
      box.setDefaultButton(qtwidgets.QMessageBox.Cancel)
      response = box.exec()
      
      if   response == qtwidgets.QMessageBox.Discard: return 0
      if   response == qtwidgets.QMessageBox.Cancel:  return 1
      elif response == qtwidgets.QMessageBox.Save:
        self.on_save()
        return self.check_save() # The user may have canceled a "save as" dialog box !
      
    else: return 0
    
  def prompt_save_filename(self):
    filename = qtwidgets.QFileDialog.getSaveFileName(self.window, editobj3.TRANSLATOR(u"Save as..."), os.path.join(onto_path[0], self.ontology.name), "OWL/XML (*.owl)")[0]
    return filename
    
  






# -*- coding: utf-8 -*-
# Owlready2
# Copyright (C) 2007-2019 Jean-Baptiste LAMY
# LIMICS (Laboratoire d'informatique médicale et d'ingénierie des connaissances en santé), UMR_S 1142
# University Paris 13, Sorbonne paris-Cité, Bobigny, France

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ["start_observing", "stop_observing", "observe", "unobserve", "isobserved", "send_event", "scan_collapsed_changes",
           "InstancesOfClass"]

import weakref

from owlready2.base import rdf_type, rdfs_subclassof, owl_equivalentclass, owl_equivalentproperty, owl_equivalentindividual
from owlready2.namespace import Ontology
from owlready2 import Thing, ThingClass

class ObservedOntology(Ontology):
  def _get_pred_value_obj(self, subject, predicate):
    if   predicate == rdf_type:
      return list(filter(None, [self._to_python(o, default_to_none = True) for o in self._get_obj_triples_sp_o(subject, predicate)]))
    else:
      return [self._to_python(o) for o in self._get_obj_triples_sp_o(subject, predicate)]
    
  def _get_pred_value_data(self, subject, predicate):
    return [self._to_python(o, d) for o, d in self._get_data_triples_sp_od(subject, predicate)]
  
  def _gen_triple_method_obj(self, triple_method):
    def f(subject, predicate, object):
      observation = self.world._observations.get(subject)
      
      #if (predicate == rdf_type) and _INSTANCES_OF_CLASS:
      #  Class = self.world._get_by_storid(object)
      #  l_2_old_instances = {}
      #  if not Class is None: # Else it is Thing
      #    for parent in Class.ancestors():
      #      for l in _INSTANCES_OF_CLASS.get(parent.storid, ()):
      #        l_2_old_instances[l] = l._get_old_value()
              
      if observation:
        #old = self._get_pred_value_obj(subject, predicate)
        triple_method(subject, predicate, object)
        #new = self._get_pred_value_obj(subject, predicate)
        observation.call(predicate)
        
      else:
        triple_method(subject, predicate, object)
        
      observation = self.world._observations.get(object)
      if observation:
        Prop = self.world._entities.get(predicate)
        if Prop and Prop.inverse:
          observation.call(Prop._inverse_storid)
          
      #if (predicate == rdf_type) and _INSTANCES_OF_CLASS and (not Class is None):
      #  for l, old_instances in l_2_old_instances.items():
      #    l._changed()
          
      if (predicate == rdf_type) and _INSTANCES_OF_CLASS:
        Class = self.world._get_by_storid(object)
        if not Class is None: # Else it is Thing
          for parent in Class.ancestors():
            for l in _INSTANCES_OF_CLASS.get(parent.storid, ()):
              l._changed()
              
    return f
  
  def _gen_triple_method_data(self, triple_method):
    def f(subject, predicate, object, datatype):
      observation = self.world._observations.get(subject)
      
      if observation:
        #old = self._get_triples_sp_od(subject, predicate)
        triple_method(subject, predicate, object, datatype)
        #new = self._get_triples_sp_od(subject, predicate)
        observation.call(predicate) #, new, old)
      else:
        triple_method(subject, predicate, object, datatype)
        
    return f
  
  def _entity_destroyed(self, entity):
    if _INSTANCES_OF_CLASS and isinstance(entity, Thing):
      Classes   = [Class for Class in entity.is_a if isinstance(Class, ThingClass)]
      Ancestors = { Ancestor for Class in Classes for Ancestor in Class.ancestors() }
      for Ancestor in Ancestors:
        for l in _INSTANCES_OF_CLASS.get(Ancestor.storid, ()):
          l._changed()

    
  
def start_observing(onto):
  if not hasattr(onto.world, "_observations"): onto.world._observations = {}
  if not onto.__class__ is ObservedOntology:
    onto.__class__ = ObservedOntology
    onto._add_obj_triple_raw_spo   = onto._gen_triple_method_obj(onto.graph._add_obj_triple_raw_spo)
    onto._set_obj_triple_raw_spo   = onto._gen_triple_method_obj(onto.graph._set_obj_triple_raw_spo)
    onto._del_obj_triple_raw_spo   = onto._gen_triple_method_obj(onto.graph._del_obj_triple_raw_spo)
    onto._add_data_triple_raw_spod = onto._gen_triple_method_data(onto.graph._add_data_triple_raw_spod)
    onto._set_data_triple_raw_spod = onto._gen_triple_method_data(onto.graph._set_data_triple_raw_spod)
    onto._del_data_triple_raw_spod = onto._gen_triple_method_data(onto.graph._del_data_triple_raw_spod)
    
def stop_observing(onto):
  onto.__class__ = Ontology
  onto._add_obj_triple_raw_spo   = onto._add_obj_triple_raw_spo
  onto._set_obj_triple_raw_spo   = onto._set_obj_triple_raw_spo
  onto._del_obj_triple_raw_spo   = onto._del_obj_triple_raw_spo
  onto._add_data_triple_raw_spod = onto._add_data_triple_raw_spod
  onto._set_data_triple_raw_spod = onto._set_data_triple_raw_spod
  onto._del_data_triple_raw_spod = onto._del_data_triple_raw_spod
  
  

_NON_EMPTY_COLLAPSED_LISTENERS = set()
class Observation(object):
  def __init__(self, o):
    self.listeners           = []
    self.collapsed_listeners = []
    self.o                   = o
    self.collapsed_changes   = set()
    
  def call(self, predicate):
    for listener in self.listeners: listener(self.o, predicate)
    
  def add_listener(self, listener, collapsed):
    if collapsed:
      if not self.collapsed_listeners:
        self.listeners.append(self.collapser)
      self.collapsed_listeners.append(listener)
    else:
      self.listeners.append(listener)
      
  def remove_listener(self, listener):
    if   listener in self.collapsed_listeners:
      self.collapsed_listeners.remove(listener)
      if not self.collapsed_listeners:
        self.listeners.remove(self.collapser)
    elif listener in self.listeners:
      self.listeners.remove(listener)
    
  def collapser(self, o, predicate):
    if not self.collapsed_changes: _NON_EMPTY_COLLAPSED_LISTENERS.add(self)
    self.collapsed_changes.add(predicate)
    
  def scan(self):
    for listener in self.collapsed_listeners: listener(self.o, self.collapsed_changes)
    self.collapsed_changes.clear()
    
    

class ObjectPack(object):
  storid = 0 # Fake storid
  
  def __init__(self, objects):
    self._objects = objects

  def __repr__(self): return "<ObjectPack %s>" % self._objects
  
  
def scan_collapsed_changes():
  for collapsed_listener in _NON_EMPTY_COLLAPSED_LISTENERS: collapsed_listener.scan()
  _NON_EMPTY_COLLAPSED_LISTENERS.clear()
  

def observe(o, listener, collapsed = False, world = None):
  if isinstance(o, ObjectPack):
    for o2 in o._objects: observe(o2, listener, collapsed, world)
    return
  
  #print("OBSERVE", sum(len(obs.collapsed_listeners) for obs in (world or o.namespace.world)._observations.values()), o, listener)
  if world is None:
    world = o.namespace.world
    o = o.storid
    
  observation = world._observations.get(o)
  if not observation: observation = world._observations[o] = Observation(o)
  observation.add_listener(listener, collapsed)
  
    
def isobserved(o, listener = None, world = None):
  if isinstance(o, ObjectPack):
    for o2 in o._objects:
      if isobserved(o2, listener, world): return True
    return False
  
  if world is None:
    world = o.namespace.world
    o = o.storid
  
  observation = world._observations.get(o)
  if listener: return observation and ((listener in observation.listeners) or (listener in observation.collapsed_listeners))
  else:        return observation and observation.listeners

def send_event(o, pred, world = None):
  if isinstance(o, ObjectPack):
    for o2 in o._objects: send_event(o2, pred)
    return
  
  if world is None:
    world = o.namespace.world
    o = o.storid
    
  observation = world._observations.get(o)
  if observation: observation.call(pred)
  
def unobserve(o, listener = None, world = None):
  if isinstance(o, ObjectPack):
    for o2 in o._objects: unobserve(o2, listener, world)
    return
  
  #print("UNOBSERVE", sum(len(obs.collapsed_listeners) for obs in (world or o.namespace.world)._observations.values()), o, listener)
  if world is None:
    world = o.namespace.world
    o = o.storid
  
  if listener:
    observation = world._observations.get(o)
    if observation:
      observation.remove_listener(listener)
      if not observation.listeners: del world._observations[o]
      
  else:
    if o in world._observations: del world._observations[o]

    
_INSTANCES_OF_CLASS = {} #weakref.WeakValueDictionary()



class StoridList(object):
  def __init__(self, graph_manager, storids):
    self.namespace = graph_manager
    self._storids = storids
    
  def _update(self): pass
      
  def __len__(self):
    if self._storids is None: self._update()
    return len(self._storids)
  
  def __iter__(self):
    if self._storids is None: self._update()
    for i in self._storids: yield self.namespace._get_by_storid(i)
    
  def __getitem__(self, i):
    if self._storids is None: self._update()
    if isinstance(i, slice):
      return [self.namespace._get_by_storid(x) for x in self._storids.__getitem__(i)]
    else:
      return self.namespace._get_by_storid(self._storids[i])
    
  def __repr__(self):
    return """<StoridList: %s>""" % list(self)
  


from owlready2.util import FirstList
class InstancesOfClass(StoridList):
  storid = 0 # Fake storid
  
  def __init__(self, Class, onto = None, order_by = "", lang = "", use_observe = False):
    self._Class         = Class
    self._lang          = lang
    self._use_observe   = use_observe
    self._Class_storids = ",".join((["'%s'" % child.storid for child in Class.descendants()]))
    
    if use_observe:
      ws = _INSTANCES_OF_CLASS.get(Class.storid)
      if ws is None: ws = _INSTANCES_OF_CLASS[Class.storid] = weakref.WeakSet()
      ws.add(self)
      
    StoridList.__init__(self, onto or Class.namespace.world, None)
    
    if order_by:
      if   isinstance(order_by, str):
        self._order_by = self.namespace.world._props[order_by].storid
      elif order_by:
        self._order_by = order_by.storid
    else:
      self._order_by = ""
      
  def __repr__(self):
    return """<InstancesOfClass "%s": %s>""" % (self._Class, list(self))
  
  def _update(self):
    if self._order_by:
      if self._lang:
        self._storids = [x[0] for x in self.namespace.graph.execute("""SELECT s FROM objs WHERE p = ? AND o IN (%s) ORDER BY (select q2.o FROM quads q2 WHERE q2.s = objs.s AND q2.p = ? AND q2.d LIKE '@%s')""" % (self._Class_storids, self._lang), (rdf_type, self._order_by))]
      else:
        self._storids = [x[0] for x in self.namespace.graph.execute("""SELECT s FROM objs WHERE p = ? AND o IN (%s) ORDER BY (select q2.o FROM quads q2 WHERE q2.s = objs.s AND q2.p = ?)""" % self._Class_storids, (rdf_type, self._order_by))]
    else:
      self._storids = [x[0] for x in self.namespace.graph.execute("""SELECT s FROM objs WHERE p = ? AND o IN (%s)""" % self._Class_storids, (rdf_type,)).fetchall()]
      
  def _get_storids(self):
    if self._storids is None: self._update()
    return self._storids
    
  def _get_old_value(self):
    if self._storids is None: self._update()
    return StoridList(self.namespace, self._storids)
  
  def _changed(self):
    observation = self.namespace.world._observations.get(self.storid)
    self._storids = None
    if observation:
      observation.call("Inverse(http://www.w3.org/1999/02/22-rdf-syntax-ns#type)")
      
  def add(self, o):
    if not self.Class in o.is_a: o.is_a.append(self.Class)
    #if not self._use_observe: self._changed()
  append = add
  
  def remove(self, o):
    destroy_entity(o)
    #self._changed()
    



# -*- coding: utf-8 -*-
# Owlready2
# Copyright (C) 2013-2019 Jean-Baptiste LAMY
# LIMICS (Laboratoire d'informatique médicale et d'ingénierie des connaissances en santé), UMR_S 1142
# University Paris 13, Sorbonne paris-Cité, Bobigny, France

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys, xml, xml.parsers.expat, urllib.parse
from collections import defaultdict

try:
  from owlready2.base import OwlReadyOntologyParsingError
except:
  class OwlReadyOntologyParsingError(Exception): pass

INT_DATATYPES   = { "http://www.w3.org/2001/XMLSchema#integer", "http://www.w3.org/2001/XMLSchema#byte", "http://www.w3.org/2001/XMLSchema#short", "http://www.w3.org/2001/XMLSchema#int", "http://www.w3.org/2001/XMLSchema#long", "http://www.w3.org/2001/XMLSchema#unsignedByte", "http://www.w3.org/2001/XMLSchema#unsignedShort", "http://www.w3.org/2001/XMLSchema#unsignedInt", "http://www.w3.org/2001/XMLSchema#unsignedLong", "http://www.w3.org/2001/XMLSchema#negativeInteger", "http://www.w3.org/2001/XMLSchema#nonNegativeInteger", "http://www.w3.org/2001/XMLSchema#positiveInteger" }
FLOAT_DATATYPES = { "http://www.w3.org/2001/XMLSchema#decimal", "http://www.w3.org/2001/XMLSchema#double", "http://www.w3.org/2001/XMLSchema#float", "http://www.w3.org/2002/07/owl#real" }


def is_bn(x):
  if isinstance(x, int): return x < 0
  return x.startswith("_")

def is_fake_bn(x):
  if isinstance(x, str): return x.startswith("_ ")
  return False

def parse(f, on_prepare_obj = None, on_prepare_data = None, new_blank = None, default_base = ""):
  parser = xml.parsers.expat.ParserCreate(None, "")
  try:
    parser.buffer_text          = True
    parser.specified_attributes = True
  except: pass
  
  stack                    = [["", ""]] # List of [parse type, value] pairs
  prefixes                 = {}
  prefixess                = [prefixes]
  tag_is_predicate         = False
  current_blank            = 0
  current_fake_blank       = 0
  current_content          = ""
  current_attrs            = None
  nb_triple                = 0
  bns                      = defaultdict(set)
  dont_create_unnamed_bn   = False
  axiom_annotation_sources = {}
  axiom_annotation_props   = {}
  axiom_annotation_targets = {}
  triples_with_unnamed_bn  = defaultdict(list)
  if default_base:
    xml_base = default_base
    if xml_base.endswith("#") or xml_base.endswith("/"): xml_base = xml_base[:-1]
    xml_dir  = xml_base.rsplit("/", 1)[0] + "/"
  else:
    xml_base                 = ""
    xml_dir                  = ""
    
  if not on_prepare_obj:
    def on_prepare_obj(s,p,o):
      nonlocal nb_triple
      nb_triple += 1
      if not is_bn(s): s = "<%s>" % s
      if not is_bn(o): o = "<%s>" % o
      print("%s %s %s ." % (s,"<%s>" % p,o))
      
    def on_prepare_data(s,p,o,d):
      nonlocal nb_triple
      nb_triple += 1
      if not is_bn(s): s = "<%s>" % s
      
      if isinstance(o, str): o = o.replace('"', '\\"').replace("\n", "\\n")
      if d and d.startswith("@"):
        print('%s %s "%s"%s .' % (s,"<%s>" % p,o,d))
      elif d:
        print('%s %s "%s"^^<%s> .' % (s,"<%s>" % p,o,d))
      else:
        print('%s %s "%s" .' % (s,"<%s>" % p,o))
        
  if not new_blank:
    def new_blank():
      nonlocal current_blank
      current_blank += 1
      return "_:%s" % current_blank
    
  def new_fake_blank():
    nonlocal current_fake_blank
    current_fake_blank += 1
    return "_ %s" % current_fake_blank
  
  node_2_blanks = defaultdict(new_blank)
  known_nodes   = set()
  
  def new_list(l):
    bn = bn0 = new_blank()
    
    if l:
      for i in range(len(l) - 1):
        on_prepare_obj(bn, "http://www.w3.org/1999/02/22-rdf-syntax-ns#first", l[i])
        bn_next = new_blank()
        on_prepare_obj(bn, "http://www.w3.org/1999/02/22-rdf-syntax-ns#rest", bn_next)
        bn = bn_next
      on_prepare_obj(bn, "http://www.w3.org/1999/02/22-rdf-syntax-ns#first", l[-1])
      on_prepare_obj(bn, "http://www.w3.org/1999/02/22-rdf-syntax-ns#rest", "http://www.w3.org/1999/02/22-rdf-syntax-ns#nil")
      
    else:
      on_prepare_obj(bn, "http://www.w3.org/1999/02/22-rdf-syntax-ns#first", "http://www.w3.org/1999/02/22-rdf-syntax-ns#nil")
      on_prepare_obj(bn, "http://www.w3.org/1999/02/22-rdf-syntax-ns#rest",  "http://www.w3.org/1999/02/22-rdf-syntax-ns#nil")
      
    return bn0
  
  def add_to_bn(bn, type, rel, value, d = ""):
    if type == "COL":
      value = tuple(frozenset(bns[v])
                    if is_bn(v) and (not v in known_nodes)
                    else v
                    for v in value)
      bns[bn].add((type, rel) + value)
    else:
      if type == "DAT":
        bns[bn].add((type, rel, value, d))
      else:
        if is_bn(value) and (not value in known_nodes): value = frozenset(bns[value])
        bns[bn].add((type, rel, value))
        
  def startNamespace(prefix, uri):
    nonlocal prefixes
    prefixes = prefixes.copy()
    prefixess.append(prefixes)
    
    if prefix: prefixes[prefix  ] = uri
    else:      prefixes[""      ] = uri
      
  def endNamespace(prefix):
    nonlocal prefixes
    prefixess.pop()
    prefixes = prefixess[-1]
    
  def startElement(tag, attrs):
    nonlocal tag_is_predicate, current_content, current_attrs, dont_create_unnamed_bn, xml_base, xml_dir
    
    tag_is_predicate = not tag_is_predicate
    if tag_is_predicate:
      
      if   attrs.get("http://www.w3.org/1999/02/22-rdf-syntax-ns#parseType") == "Collection":
        stack.append(["Collection", []])
        
      elif tag == "http://www.w3.org/1999/02/22-rdf-syntax-ns#RDF":
        stack.append(["RDF", ""])
        
        namespace_base = attrs.get("http://www.w3.org/XML/1998/namespacebase")

        if namespace_base:
          xml_base = namespace_base
          if "/" in namespace_base:
            xml_dir  = namespace_base.rsplit("/", 1)[0] + "/"
            
      else:
        iri = attrs.get("http://www.w3.org/1999/02/22-rdf-syntax-ns#resource")
        if iri:
          if   iri.startswith("#"): iri = xml_base + iri
          elif iri.startswith("/"): iri = xml_dir  + iri[1:]
          elif not iri:             iri = xml_base
          elif not ":" in iri:      iri = urllib.parse.urljoin(xml_dir, iri)
          if iri.endswith("/"): iri = iri[:-1]
          stack.append(["Resource", iri])
          
        else:
          iri = attrs.get("http://www.w3.org/1999/02/22-rdf-syntax-ns#nodeID")
          
          if iri:
            iri = node_2_blanks[iri]
            known_nodes.add(iri)
            stack.append(["Resource", iri])
          else:
            stack.append(["Literal", ""])
            current_content = ""
            current_attrs   = attrs
            
          if (tag == "http://www.w3.org/2002/07/owl#annotatedSource") or (tag == "http://www.w3.org/2002/07/owl#annotatedTarget"):
            dont_create_unnamed_bn = True
            
    else:
      iri = attrs.get("http://www.w3.org/1999/02/22-rdf-syntax-ns#about", None)
      if iri is None:
        iri = attrs.get("http://www.w3.org/1999/02/22-rdf-syntax-ns#ID", None)
        if iri: iri = "#%s" % iri
      else:
        if iri.endswith("/"): iri = iri[:-1]
      if iri is None:
        iri = attrs.get("http://www.w3.org/1999/02/22-rdf-syntax-ns#nodeID")
        if iri:
          iri = node_2_blanks[iri]
          known_nodes.add(iri)
        else:
          if dont_create_unnamed_bn: iri = new_fake_blank()
          else:                      iri = new_blank()
      else:
        if   iri.startswith("#"): iri = xml_base + iri
        elif iri.startswith("/"): iri = xml_dir  + iri[1:]
        elif not iri:             iri = xml_base
        elif not ":" in iri:      iri = urllib.parse.urljoin(xml_dir, iri)
          
      if tag != "http://www.w3.org/1999/02/22-rdf-syntax-ns#Description":
        if not is_fake_bn(iri):
          on_prepare_obj(iri, "http://www.w3.org/1999/02/22-rdf-syntax-ns#type", tag)
        if is_bn(iri):
          add_to_bn(iri, "REL", "http://www.w3.org/1999/02/22-rdf-syntax-ns#type", tag)
          
      if stack[-1][0] == "Collection":
        stack[-1][1].append(iri)
        
      else:
        if stack[-1][0] == "Literal": stack[-1][0] = "Resource"
        stack[-1][1] = iri
        
        
  def endElement(tag):
    nonlocal tag_is_predicate, dont_create_unnamed_bn
    
    if tag_is_predicate:
      parse_type, value = stack.pop()
      
      if stack[-1][0] == "Collection": iri = stack[-1][1][-1]
      else:                            iri = stack[-1][1]
      
      if   tag == "http://www.w3.org/2002/07/owl#annotatedSource":
        dont_create_unnamed_bn = False
        axiom_annotation_sources[iri] = value
        if is_fake_bn(value):
          triples_with_unnamed_bn[iri].insert(0, (tag, value, parser.CurrentLineNumber, parser.CurrentColumnNumber))
          
          tag_is_predicate = not tag_is_predicate
          return
        
      elif tag == "http://www.w3.org/2002/07/owl#annotatedProperty":
        axiom_annotation_props  [iri] = value
      
      elif tag == "http://www.w3.org/2002/07/owl#annotatedTarget":
        dont_create_unnamed_bn = False
        axiom_annotation_targets[iri] = value
        if is_fake_bn(value):
          triples_with_unnamed_bn[iri].append((tag, value, parser.CurrentLineNumber, parser.CurrentColumnNumber))
          
          tag_is_predicate = not tag_is_predicate
          return
        
      
      if   parse_type == "Resource":
        if not is_fake_bn(iri):
          on_prepare_obj(iri, tag, value)
          
        if is_bn(iri):
          add_to_bn(iri, "REL", tag, value)
            
        if is_bn(value):
          add_to_bn(value, "INV", tag, iri)
          
          
      elif parse_type == "Literal":
        o = current_content
        d = current_attrs.get("http://www.w3.org/XML/1998/namespacelang")
        if d is None:
          d = current_attrs.get("http://www.w3.org/1999/02/22-rdf-syntax-ns#datatype", "")
          if   d in INT_DATATYPES:   o = int  (o)
          elif d in FLOAT_DATATYPES: o = float(o)
        else:
          d = "@%s" % d
          
        if not is_fake_bn(iri):
          on_prepare_data(iri, tag, o, d)
        if is_bn(iri):
          add_to_bn(iri, "DAT", tag, o, d)
          
      elif parse_type == "Collection":
        if not is_fake_bn(iri):
          on_prepare_obj(iri, tag, new_list(value))
        if is_bn(iri):
          add_to_bn(iri, "COL", tag, value)
          
          
    tag_is_predicate = not tag_is_predicate
    
    
  def characters(content):
    nonlocal current_content
    if stack[-1][0] == "Literal": current_content += content
    
    
  parser.StartNamespaceDeclHandler = startNamespace
  parser.EndNamespaceDeclHandler   = endNamespace
  parser.StartElementHandler       = startElement
  parser.EndElementHandler         = endElement
  parser.CharacterDataHandler      = characters
  
  try:
    if isinstance(f, str):
      f = open(f, "rb")
      parser.ParseFile(f)
      f.close()
    else:
      parser.ParseFile(f)
  except Exception as e:
    raise OwlReadyOntologyParsingError("RDF/XML parsing error in file %s, line %s, column %s." % (getattr(f, "name", getattr(f, "url", "???")), parser.CurrentLineNumber, parser.CurrentColumnNumber)) from e
  
  
  if triples_with_unnamed_bn:
    content_2_bns = defaultdict(list)
    for bn, content in bns.items():
      if not is_fake_bn(bn):
        content_2_bns[frozenset(content)].append(bn)
        
    def rebuild_bn(content):
      bn = new_blank()
      content_2_bns[frozenset(content)].append(bn)
      for i in content:
        if   i[0] == "REL":
          drop, p, o = i
          if not isinstance(o, (str, int)): o = rebuild_bn(o)
          on_prepare_obj(bn, p, o)
        elif i[0] == "DAT":
          drop, p, o, d = i
          if not isinstance(o, (str, int)): o = rebuild_bn(o)
          on_prepare_data(bn, p, o, d)
        elif i[0] == "INV":
          drop, p, o = i
          if not isinstance(o, (str, int)): o = rebuild_bn(o)
          on_prepare_obj(o, p, bn)
        elif i[0] == "COL":
          drop, p, *l = i
          l = [(isinstance(x, (str, int)) and x) or rebuild_bn(x) for x in l]
          o = new_list(l)
          on_prepare_obj(bn, p, o)
        else:
          print(i)
          raise ValueError
      return bn
    
    for axiom_iri, triples in triples_with_unnamed_bn.items():
      for p, o, line, column in triples:
        try:
          content = bns[o]
          if p == "http://www.w3.org/2002/07/owl#annotatedSource":
            target = axiom_annotation_targets[axiom_iri]
            if is_bn(target): target = frozenset(bns[target])
            candidates_bn = content_2_bns[frozenset(content | { ("REL", axiom_annotation_props[axiom_iri], target) })]
            
          else:
            source = axiom_annotation_sources[axiom_iri]
            if is_bn(source):
              source = frozenset(bns[source] | { ("REL", axiom_annotation_props[axiom_iri], target) })
            candidates_bn = (content_2_bns[frozenset(content | { ("INV", axiom_annotation_props[axiom_iri], source) })] or
                             content_2_bns[frozenset(content)])
            
          if candidates_bn: o = candidates_bn[-1]
          else:
            o = rebuild_bn(content)
          on_prepare_obj(axiom_iri,p,o)
        except Exception as e:
          raise OwlReadyOntologyParsingError("RDF/XML parsing error in file %s, line %s, column %s." % (getattr(f, "name", getattr(f, "url", "???")), line, column)) from e
        
    
  return nb_triple


if __name__ == "__main__":
  filename = sys.argv[-1]
  
  import time
  t = time.time()
  nb_triple = parse(filename)
  t = time.time() - t
  print("# %s triples read in %ss" % (nb_triple, t), file = sys.stderr)



# Owlready2
# Copyright (C) 2021 Jean-Baptiste LAMY
# LIMICS (Laboratoire d'informatique médicale et d'ingénierie des connaissances en santé), UMR_S 1142
# University Paris 13, Sorbonne paris-Cité, Bobigny, France

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import sys, os, re
from owlready2 import *

"""

Not supported:

ASK
DESCRIBE
LOAD, ADD, MOVE, COPY, CLEAR, DROP

GRAPH, FROM, FROM NAMED

CONSTRUCT

SERVICE

INSERT DATA, DELETE DATA, DELETE WHERE (use INSERT or DELETE instead)

VALUES

MINUS

"""

from owlready2.sparql.parser import *
from owlready2.sparql.func   import register_python_function, FuncSupport

_RE_AUTOMATIC_INDEX = re.compile(r"TABLE (.*?) AS (.*?) USING AUTOMATIC.*\((.*?)\)")
_RE_NORMAL_INDEX    = re.compile(r"TABLE (.*?) AS (.*?) USING (COVERING )?INDEX (.*?) ")

_DEPRIORIZE_SUBQUERIES_OPT = True

class Translator(object):
  def __init__(self, world, error_on_undefined_entities = True):
    self.world                         = world
    self.error_on_undefined_entities   = error_on_undefined_entities
    self.prefixes                      = { "rdf:" : "http://www.w3.org/1999/02/22-rdf-syntax-ns#", "rdfs:" : "http://www.w3.org/2000/01/rdf-schema#", "owl:" : "http://www.w3.org/2002/07/owl#", "xsd:" : "http://www.w3.org/2001/XMLSchema#", "obo:" : "http://purl.obolibrary.org/obo/", "owlready:" : "http://www.lesfleursdunormal.fr/static/_downloads/owlready_ontology.owl#" }
    self.base_iri                      = ""
    self.current_anonynous_var         = 0
    self.current_parameter             = 0
    self.main_query                    = None
    self.preliminary_selects           = []
    self.recursive_preliminary_selects = {}
    self.escape_mark                   = "@@@ESCAPE@@@"
    
    if not getattr(world.graph, "_has_sparql_func", False):
      register_python_function(world)
      world.graph._has_sparql_func = True
      
  def parse(self, sparql):
    while self.escape_mark in sparql:
      self.escape_mark += "ç"
    CURRENT_TRANSLATOR.set(self)
    self.main_query = PARSER.parse(LEXER.lex(sparql))
    
    sql = ""
    if self.preliminary_selects:
      sql += "WITH "
      if max(prelim.recursive for prelim in self.preliminary_selects): sql += "RECURSIVE "
      sql += ",\n\n".join(prelim.sql() for prelim in self.preliminary_selects)
      sql += "\n\n"
    sql += self.main_query.sql()
    
    if self.solution_modifier[0]:
      sql += " GROUP BY %s" % ", ".join(self.main_query.parse_expression(x) for x in self.solution_modifier[0])
    if self.solution_modifier[1]:
      sql += " HAVING %s" % self.main_query.parse_expression(self.solution_modifier[1])
    if self.solution_modifier[2]:
      sql += " ORDER BY %s" % ", ".join(self.main_query.parse_expression(x) for x in self.solution_modifier[2])
    if self.solution_modifier[3]:
      sql += " LIMIT %s"  % self._to_sql(self.solution_modifier[3])
    if self.solution_modifier[4]:
      if not self.solution_modifier[3]: sql += " LIMIT -1" # SQLite requires a LIMIT clause before the OFFSET clause
      sql += " OFFSET %s" % self._to_sql(self.solution_modifier[4])
      
    nb_parameter = self.current_parameter
    parameter_2_parameter_datatypes = {}
    parameter_datatypes = []
    if self.escape_mark in sql:
      def sub(m):
        escape = m.group(0)[len(self.escape_mark):]
        if escape.startswith("TypeOfParam?"):
          number = int(escape[12:])
          r = parameter_2_parameter_datatypes.get(number)
          if r is None:
            r = parameter_2_parameter_datatypes[number] = self.new_parameter()
            parameter_datatypes.append(number - 1)
        return "?%s" % r
      sql = re.sub("%s[^ ]*" % self.escape_mark, sub, sql)

    #sql = self.optimize_sql(sql)
    
    if   self.main_query.type == "select":
      return PreparedSelectQuery(self.world, sql, [column.var for column in self.main_query.columns if not column.name.endswith("d")], [column.type for column in self.main_query.columns], nb_parameter, parameter_datatypes)
    
    elif self.main_query.type == "modify":
      select_param_indexes = [i - 1 for i in self.main_query.select_param_indexes]
      return PreparedModifyQuery(self.world, sql, [column.var for column in self.main_query.columns if not column.name.endswith("d")], [column.type for column in self.main_query.columns], nb_parameter, parameter_datatypes, self.world.get_ontology(self.main_query.ontology_iri.value) if self.main_query.ontology_iri else None, self.parse_inserts_deletes(self.main_query.deletes, self.main_query.columns), self.parse_inserts_deletes(self.main_query.inserts, self.main_query.columns), select_param_indexes)
    
    
  def optimize_sql(self, sql):
    # Avoid Sqlite3 AUTOMATIC INDEX when a similar index can be used.
    plan = list(self.world.graph.execute("""EXPLAIN QUERY PLAN %s""" % sql))
    #for l in plan:
      #if (" USING AUTOMATIC " in l[3]) and not (" TABLE " in l[3]): break
    #else:
    try:
      self.world.graph.execute("PRAGMA automatic_index = false")
      plan = list(self.world.graph.execute("""EXPLAIN QUERY PLAN %s""" % sql))
    finally:
      self.world.graph.execute("PRAGMA automatic_index = true")
      
    for l in plan:
        match = _RE_NORMAL_INDEX.search(l[3])
        if match:
          table_type = match.group(1)
          table_name = match.group(2)
          index_name = match.group(4)
          if table_name == "q": continue # Recursive hard-coded preliminary queries
          #if ("%s INDEXED" % table_name) in sql: continue
          table_def = "%s %s" % (table_type, table_name)
          #print("OPTIMIZE!!!", l, table_type, table_name, index_name)
          sql = sql.replace(table_def, "%s INDEXED BY %s" % (table_def, index_name), 1)
    return sql
    
    
    # while True:
    #   changed = False
    #   plan = list(self.world.graph.execute("""EXPLAIN QUERY PLAN %s""" % sql))
      
    #   for l in plan:
    #     match = _RE_AUTOMATIC_INDEX.search(l[3])
    #     if match:
    #       table_type = match.group(1)
    #       table_name = match.group(2)
    #       #if table_name in indexeds: continue
    #       cols       = { i[0] for i in match.group(3).split(" AND ") }
          
    #       if cols == { "s", "p", "o" }:
    #         #print("OPTIMIZE!!!", l, table_type, table_name, cols)
    #         table_def = "%s %s" % (table_type, table_name)
    #         sql_s = sql.replace(table_def, "%s INDEXED BY index_%s_sp" % (table_def, table_type), 1)
    #         sql_o = sql.replace(table_def, "%s INDEXED BY index_%s_op" % (table_def, table_type), 1)
    #         sql_s = self.optimize_sql(sql_s)
    #         sql_o = self.optimize_sql(sql_o)
    #         s_nb_not_indexed = sql_s.count("NOT INDEXED")
    #         o_nb_not_indexed = sql_o.count("NOT INDEXED")
    #         if s_nb_not_indexed <= o_nb_not_indexed: sql = sql_s #; print("CHOIX s", s_nb_not_indexed, o_nb_not_indexed)
    #         else:                                    sql = sql_o #; print("CHOIX o", s_nb_not_indexed, o_nb_not_indexed)
    #         #indexeds.add(table_name)
    #         changed = True
    #         break
          
    #   if changed: continue
            
    #   for l in plan:
    #     match = _RE_AUTOMATIC_INDEX.search(l[3])
    #     if match:
    #       table_type = match.group(1)
    #       table_name = match.group(2)
    #       cols       = { i[0] for i in match.group(3).split(" AND ") }
    #       #print("OPTIMIZE!!!", l, table_type, table_name, cols)
          
    #       table_def = "%s %s" % (table_type, table_name)
    #       if   "s" in cols:
    #         sql = sql.replace(table_def, "%s INDEXED BY index_%s_sp" % (table_def, table_type), 1)
    #         #indexeds.add(table_name)
    #         changed = True
    #       elif "o" in cols:
    #         sql = sql.replace(table_def, "%s INDEXED BY index_%s_op" % (table_def, table_type), 1)
    #         #indexeds.add(table_name)
    #         changed = True
    #       elif "p" in cols:
    #         sql = sql.replace(table_def, "%s NOT INDEXED" % table_def, 1)
    #         #indexeds.add(table_name)
    #         changed = True
            
    #   if not changed: break
    # return sql




  
  def parse_inserts_deletes(self, triples, columns):
    var_2_column = { column.var : column for column in self.main_query.columns if not column.name.endswith("d") }
    r = []
    for triple0 in triples:
      triple = []
      for x in triple0:
        if   hasattr(x, "storid"):
          triple.append(("objs", x.storid))
        elif x.name == "VAR":
          if x.value.startswith("??anon") or x.value.startswith("_:"): # a new blank node
            triple.append(("bn", x.value))
          else: # a normal var
            column = var_2_column[x.value]
            triple.append(("vars", column.index))
            if len(triple) == 3: # in 'o' position => can be objs or datas!
              if column.index + 1 < len(columns):
                next_column = columns[column.index]
                if next_column.name.endswith("d"):
                  triple.append(("vars", next_column.index))
        elif x.name == "PARAM":
          triple.append(("param", x.number - 1))
          if len(triple) == 3: # in 'o' position => can be datas!
            triple.append(("paramdatatype", x.number - 1))
        else:
          if   isinstance(x.value, locstr):
            triple.append(("datas", x.value[1:-1]))
            triple.append(("datas", "@%s" % x.value.lang))
          else:
            if isinstance(x.value, str): v, d = self.world._to_rdf(x.value[1:-1])
            else:                        v, d = self.world._to_rdf(x.value)
            d2 = getattr(x, "datatype", None) or d
            triple.append(("datas", v))
            triple.append(("datas", d2))
            
      r.append(triple)
    return r
  
  def new_sql_query(self, name, block, selects = None, distinct = None, solution_modifier = None, preliminary = False, extra_binds = None, nested_inside = None, copy_vars = False):
    if preliminary and not name: name = "prelim%s" % (len(self.preliminary_selects) + 1)
    
    #print("BLOCK", name, block)
    
    if isinstance(block, UnionBlock) and block.simple_union_triples:
      block = SimpleTripleBlock(block.simple_union_triples)
      
    if   isinstance(block, SimpleTripleBlock):
      s = SQLQuery(name)
      
    elif isinstance(block, OptionalBlock):
      s = SQLQuery(name)
      s.optional = True
      
    elif isinstance(block, UnionBlock):
      s = SQLCompoundQuery(name, nested_inside)
      if selects is None: selects = block.get_ordered_vars()
      
    elif isinstance(block, FilterBlock):
      s = SQLNestedQuery(name)
      #if isinstance(block, ExistsBlock): s.extra_sql = "IS NOT NULL"
      #else:                              s.extra_sql = "IS NULL"
      s.exists = isinstance(block, ExistsBlock)
      s.next_table_id = nested_inside.next_table_id
      s.vars = nested_inside.vars
      preliminary = False
      
    elif isinstance(block, NotExistsBlock):
      s = SQLCompoundQuery(name, nested_inside)
      
    elif isinstance(block, SubQueryBlock):
      s = block.parse()
      s.name = "prelim%s" % (len(self.preliminary_selects) + 1)
      
    else:
      raise ValueError("Unknown block type '%s'!" % block)
    
    if preliminary:
      s.preliminary = True
      self.preliminary_selects.append(s)
      
      
    # if copy_vars:
    #   if hasattr(nested_inside, "vars"):
    #     s.vars = nested_inside.vars.copy()
    #   elif hasattr(nested_inside, "parent") and nested_inside.parent:
    #     s.vars = nested_inside.parent.vars.copy()

        
    extra_binds = extra_binds or []
    if isinstance(selects, list): # Otherwise, it is "SELECT *"
      for i, select in enumerate(selects):
        if isinstance(select, list): # ( expression AS var )
          selects[i] = select[2]
          bind = Bind(select[0], select[2])
          extra_binds.append(bind)
          
          
    if   isinstance(block, (SimpleTripleBlock, OptionalBlock, FilterBlock)):
      s.parse_distinct(distinct)
      
      for i in block:
        if isinstance(i, Bind): s.prepare_bind(i)
      for bind in extra_binds: s.prepare_bind(bind)
      
      s.parse_selects(selects)
      
      s.parse_triples(block)
      for bind in extra_binds: s.parse_bind(bind)
      
      s.finalize_columns()
      
    elif isinstance(block, UnionBlock):
      for alternative in block:
        #print("   ", s.parent.vars)
        query = self.new_sql_query(None, alternative, selects, distinct, None, False, extra_binds, nested_inside = s, copy_vars = False)
        s.append(query, "UNION")
      s.finalize_compounds()
      
      
    if isinstance(block, FilterBlock):
      nested_inside.next_table_id = s.next_table_id
      
    if (not preliminary) and solution_modifier: self.solution_modifier = solution_modifier
    return s

  def expand_prefix(self, prefix):
    expanded = self.prefixes.get(prefix)
    if expanded: return expanded
    prefix0 = prefix[:-1] # Remove trailing :
    for ontology in self.world.ontologies.values():
      if prefix0 == ontology.name:
        self.prefixes[prefix] = ontology.base_iri
        return ontology.base_iri
    raise ValueError("Undefined prefix '%s'!" % prefix)

  def abbreviate(self, entity):
    if self.error_on_undefined_entities:
      r = self.world._abbreviate(entity, False)
      if r is None: raise ValueError("No existing entity for IRI '%s'! (use error_on_undefined_entities=False to accept unknown entities in SPARQL queries)" % entity)
      return r
    else:
      return self.world._abbreviate(entity)
  
  def _to_sql(self, x):
    if x.name == "PARAM": return "?%s" % x.number
    return x.value
    
  def new_var(self):
    self.current_anonynous_var += 1
    return "??anon%s" % self.current_anonynous_var

  def new_parameter(self):
    self.current_parameter += 1
    return self.current_parameter
  
  def get_recursive_preliminary_select(self, triple, fixed, fixed_var, prelim_triples):
    prelim = self.recursive_preliminary_selects.get((triple, fixed, fixed_var, tuple(prelim_triples)))
    if not prelim:
      self.recursive_preliminary_selects[triple, fixed, fixed_var, tuple(prelim_triples)] = prelim = SQLRecursivePreliminaryQuery("prelim%s" % (len(self.preliminary_selects) + 1), triple, fixed, fixed_var)
      self.preliminary_selects.append(prelim)
      prelim.build(triple, prelim_triples)
    return prelim


class PreparedQuery(object):
  def __init__(self, world, sql, column_names, column_types, nb_parameter, parameter_datatypes):
    self.world               = world
    self.sql                 = sql
    self.column_names        = column_names
    self.column_types        = column_types
    self.nb_parameter        = nb_parameter
    self.parameter_datatypes = parameter_datatypes
  
  def execute(self, params = ()):
    self.world._nb_sparql_call += 1
    sql_params = [self.world._to_rdf(param)[0] for param in params]
    for i in self.parameter_datatypes: sql_params.append(self.world._to_rdf(params[i])[1])
    return self.world.graph.execute(self.sql, sql_params)
  
class PreparedSelectQuery(PreparedQuery):
  def execute(self, params = ()):

    for l in PreparedQuery.execute(self, params):
      l2 = []
      i = 0
      while i < len(l):
        if self.column_types[i] == "objs":
          if l[i] is None: l2.append(None)
          else: l2.append(self.world._to_python(l[i], None) or l[i])
          i += 1
        else:
          if l[i + 1] is None:
            if l[i] is None: l2.append(None)
            else:            l2.append(self.world._to_python(l[i], None) or l[i])
          else:
            l2.append(self.world._to_python(l[i], l[i + 1]))
          i += 2
      yield l2
      
  def execute_csv(self, params = (), separator = ","):
    import csv, io
    b = io.StringIO()
    f = csv.writer(b, delimiter = separator)
    f.writerow(col[1:] for col in self.column_names)
    rows = []

    for l in PreparedQuery.execute(self, params):
      l2 = []
      i = 0
      while i < len(l):
        if self.column_types[i] == "objs":
          if   l[i] is None: l2.append("")
          elif l[i] > 0:     l2.append(self.world._unabbreviate(l[i]))
          else:              l2.append("_:%s" % (-l[i]))
          i += 1
        else:
          if l[i + 1] is None:
            if   l[i] is None: l2.append("")
            elif l[i] > 0:     l2.append(self.world._unabbreviate(l[i]))
            else:              l2.append("_:%s" % (-l[i]))
          else:
            l2.append(str(self.world._to_python(l[i], l[i + 1])))
          i += 2
      f.writerow(l2)
    return b.getvalue()
  
  def execute_tsv(self, params = ()): return self.execute_csv(params, "\t")

  def execute_json(self, params = ()):
    bindings = []
    colnames = [col[1:] for col in self.column_names]
    json = { "head" : { "vars" : colnames },
             "results" : { "bindings" : bindings } }
    for l in PreparedQuery.execute(self, params):
      binding = {}
      bindings.append(binding)
      i = 0
      c = 0
      while i < len(l):
        if self.column_types[i] == "objs":
          if   l[i] is None: pass
          elif l[i] > 0:     binding[colnames[c]] = { "type" : "uri", "value" : self.world._unabbreviate(l[i]) }
          else:              binding[colnames[c]] = { "type" : "bnode", "value" : "r%s" % (-l[i]) }
          i += 1
        else:
          if l[i + 1] is None:
            if   l[i] is None: pass
            elif l[i] > 0:     binding[colnames[c]] = { "type" : "uri", "value" : self.world._unabbreviate(l[i]) }
            else:              binding[colnames[c]] = { "type" : "bnode", "value" : "r%s" % (-l[i]) }
          else:
            value = str(self.world._to_python(l[i], l[i + 1]))
            if   isinstance(l[i + 1], str): binding[colnames[c]] = { "type" : "literal", "value" : value, "xml:lang" : l[i + 1][1:] }
            elif l[i + 1]:                  binding[colnames[c]] = { "type" : "literal", "value" : value, "datatype" : self.world._unabbreviate(l[i + 1]) }
            else:                           binding[colnames[c]] = { "type" : "literal", "value" : value }
          i += 2
        c += 1
    return repr(json)

  def execute_xml(self, params = ()):
    bindings = []
    colnames = [col[1:] for col in self.column_names]
    xml = """<?xml version="1.0"?>
<sparql xmlns="http://www.w3.org/2005/sparql-results#">
  <head>
"""
    for colname in colnames:
      xml += """    <variable name="%s"/>\n""" % colname
    xml += """  </head>
  <results>
"""
    
    for l in PreparedQuery.execute(self, params):
      xml += """    <result>\n"""
      i = 0
      c = 0
      while i < len(l):
        xml += """      <binding name="%s">\n""" % colnames[c]
        if self.column_types[i] == "objs":
          if   l[i] is None: pass
          elif l[i] > 0:     xml += """        <uri>%s</uri>\n""" % self.world._unabbreviate(l[i])
          else:              xml += """        <bnode>r%s</bnode>\n""" % (-l[i])
          i += 1
        else:
          if l[i + 1] is None:
            if   l[i] is None: pass
            elif l[i] > 0:     xml += """        <uri>%s</uri>\n""" % self.world._unabbreviate(l[i])
            else:              xml += """        <bnode>r%s</bnode>\n""" % (-l[i])
          else:
            value = str(self.world._to_python(l[i], l[i + 1]))
            if   isinstance(l[i + 1], str): xml += """        <literal xml:lang="%s">%s</literal>\n""" % (l[i + 1][1:], value)
            elif l[i + 1]:                  xml += """        <literal datatype="%s">%s</literal>\n""" % (self.world._unabbreviate(l[i + 1]), value)
            else:                           xml += """        <literal>%s</literal>\n""" % value
          i += 2
        c += 1
        xml += """      </binding>\n"""
      xml += """    </result>\n"""
        
    xml += """  </results>
</sparql>
"""
    return xml
  
  
class PreparedModifyQuery(PreparedQuery):
  def __init__(self, world, sql, column_names, column_types, nb_parameter, parameter_datatypes, ontology, deletes, inserts, select_param_indexes):
    PreparedQuery.__init__(self, world, sql, column_names, column_types, nb_parameter, parameter_datatypes)
    
    column_name_2_index = { column_name : i for (i, column_name) in enumerate(column_names) }
    self.ontology = ontology
    self.deletes  = deletes
    self.inserts  = inserts
    self.select_param_indexes = select_param_indexes
    
  def execute(self, params = ()):
    nb_match = 0
    if self.sql: resultss = PreparedQuery.execute(self, [params[i] for i in self.select_param_indexes])
    else:        resultss = [[]]
    for results in list(resultss):
      nb_match += 1
      
      for delete in self.deletes:
        triple = []
        for type, value in delete:
          if   type == "vars":          triple.append(results[value])
          elif type == "param":         triple.append(self.world._to_rdf(params[value])[0])
          elif type == "paramdatatype": triple.append(self.world._to_rdf(params[value])[1])
          else:                         triple.append(value)
        #print("DEL", triple)
        self.world._del_triple_with_update(*triple)
        
      bns = {}
      for insert in self.inserts:
        triple = []
        for type, value in insert:
          if   type == "vars":          triple.append(results[value])
          elif type == "bn":            triple.append(bns.get(value) or bns.setdefault(value, self.world.new_blank_node()))
          elif type == "param":         triple.append(self.world._to_rdf(params[value])[0])
          elif type == "paramdatatype": triple.append(self.world._to_rdf(params[value])[1])
          else:                         triple.append(value)
        #print("ADD", triple)
        self.world._add_triple_with_update(self.ontology, *triple)
        
    return nb_match
  
    
class Column(object):
  def __init__(self, var, type, binding, name, index):
    self.var         = var
    self.type        = type
    self.binding     = binding
    self.name        = name
    self.index       = index
    
  def __repr__(self):
    return """<Column #%s %s %s %s %s>""" % (self.index, self.var, self.type, self.binding, self.name)
    
class Variable(object):
  def __init__(self, name):
    self.name           = name
    self.type           = "quads"
    self.fixed_datatype = None
    self.prop_type      = "quads" # Type when the variable is used as a property
    self.bindings       = []
    self.bind           = None
    self.initial_query  = None
    self.nb_table       = 0
    
  def __repr__(self): return """<Variable %s type %s, %s bindings>""" % (self.name, self.type, len(self.bindings))
  
  def get_binding(self, query):
    if not self.bindings:
      #print("* Owlready2 * WARNING: variable without binding in SPARQL, use a suboptimal option", file = sys.stderr)
      table = Table("any%s" % query.next_table_id, """(SELECT DISTINCT s AS o, NULL AS d FROM quads UNION SELECT DISTINCT o, d FROM quads)""")
      table.query = query
      query.tables.append(table)
      query.name_2_table[table.name] = table
      query.next_table_id += 1
      self.bindings.append("%s.o" % table.name)
    i = 0
    for binding in self.bindings:
      if not binding.startswith("IN "): break
    return binding
  
  def update_type(self, type):
    if   self.type == "quads": self.type = type
    elif (type != self.type) and (type != "quads") :
      raise ValueError("Variable %s cannot be both %s and %s!" % (self.name, self.type, type))
    
class Table(object):
  def __init__(self, name, type = "quads", index = None, join = ",", join_conditions = None):
    self.name            = name
    self.type            = type
    self.index           = index
    self.join            = join
    self.join_conditions = join_conditions or []
    self.query           = None
    
  def __repr__(self): return "<Table '%s %s'>" % (self.type, self.name)
    
  def sql(self):
    return """%s %s%s%s""" % (self.type, self.name, self.index and (" INDEXED BY %s" % self.index) or "", self.join_conditions and (" ON (%s)" % " AND ". join(self.join_conditions)) or "")
  
  
class SQLQuery(FuncSupport):
  def __init__(self, name):
    self.name                     = name
    self.preliminary              = False
    self.recursive                = False
    self.translator               = CURRENT_TRANSLATOR.get()
    self.distinct                 = False
    self.raw_selects              = None
    self.next_table_id            = 1
    self.tables                   = []
    self.name_2_table             = {}
    self.columns                  = []
    self.conditions               = []
    self.triples                  = []
    self.vars_needed_for_select   = set()
    self.vars                     = {}
    self.extra_sql                = ""
    self.select_simple_union      = False
    self.optional                 = False
    self.has_one                  = False
    
  def __repr__(self): return "<%s '%s'>" % (self.__class__.__name__, self.sql())
  
  def sql(self):
    if self.tables:
      sql = """SELECT """
      if self.distinct: sql += "DISTINCT "
      sql += """%s FROM """ % (", ".join(str(column.binding) for column in self.columns))
      for table in self.tables:
        if not table is self.tables[0]: sql += " %s " % table.join
        sql += table.sql()
      if self.conditions:
        sql += """ WHERE %s""" % (" AND ".join(str(condition) for condition in self.conditions))
    else:
      if not self.columns: return ""
      if self.select_simple_union:
        for i, column in enumerate(self.columns):
          if isinstance(column.binding, list): break
        l = []
        for j in range(len(self.columns[i].binding)):
          l.append([])
          for k, column in enumerate(self.columns):
            if k == i: l[-1].append(column.binding[j])
            else:      l[-1].append(column.binding)
        sql = """VALUES %s""" % ",".join("(%s)" % ",".join(str(k) for k in j) for j in l)
      else:
        sql = """VALUES (%s)""" % (",".join(str(column.binding) for column in self.columns))
        
    if self.extra_sql: sql += " %s" % self.extra_sql
    if self.preliminary: return """%s(%s) AS (%s)""" % (self.name, ", ".join(column.name for column in self.columns), sql)
    return sql
    
  def set_column_names(self, names):
    for column, name in zip(self.columns, names): column.name = name
    
  def parse_distinct(self, distinct):
    if isinstance(distinct, rply.Token): self.distinct = distinct and (distinct.value.upper() == "DISTINCT")
    else:                                self.distinct = distinct
    
  def parse_var(self, x):
    if isinstance(x, rply.Token): name = x.value
    else:                         name = x
    var = self.vars.get(name)
    if not var: self.vars[name] = var = Variable(name)
    return var
  
  def parse_selects(self, selects):
    if selects is None:
      self.raw_selects = "*"
    else:
      self.raw_selects = selects
      vars_needed_for_select = { self.parse_var(select) for select in selects if (isinstance(select, rply.Token) and (select.name == "VAR")) or (isinstance(select, str)) }
      for var in vars_needed_for_select:
        self.expand_referenced_vars(var, self.vars_needed_for_select)
        
  def expand_referenced_vars(self, var, r):
    r.add(var)
    if var.bind:
      for var_name in var.bind.referenced_var_names:
        self.expand_referenced_vars(self.parse_var(var_name), r)
        
  def prepare_bind(self, bind):
    var      = self.parse_var(bind.var)
    var.bind = bind
    
  def parse_bind(self, bind):
    var      = self.parse_var(bind.var)
    var.bind = bind
    var.bindings.insert(0, self.parse_expression(bind.expression))
    
    fixed_type, fixed_datatype = self.infer_expression_type(bind.expression, accept_zero = True)
    if fixed_type is None: fixed_type = "quads"
    var.update_type(fixed_type)
    if fixed_type != "objs":  var.fixed_datatype = fixed_datatype
    
  def parse_filter(self, filter):
    sql = self.parse_expression(filter.constraint)
    self.conditions.append(sql)
    
  def add_subquery(self, sub):
    if isinstance(sub, SQLNestedQuery):
      self.conditions.append(sub)
      if _DEPRIORIZE_SUBQUERIES_OPT and not self.has_one:
        self.tables.append(Table("one", "one"))
        self.has_one = True
    else:
      table = Table("p%s" % self.next_table_id, sub.name)
      table.query = sub
      self.next_table_id += 1
      self.tables.append(table)
      self.name_2_table[table.name] = table
      if sub.optional:
        table.join = "LEFT JOIN"
        conditions = table.join_conditions
      else:
        conditions = self.conditions

      for column in sub.columns:
        var = self.parse_var(column.var)
        var.update_type(column.type)
        if not column.name.endswith("d"):
          self.create_conditions(conditions, table, column.name, var)
          
          
  def parse_triples(self, triples):
    if self.triples: raise ValueError("Cannot parse triples twice!")
    self.block = triples
    self.triples.extend(triples)
    
    if self.raw_selects is None: raise ValueError("Need to call parse_selects() before finalizing triples!")
    
    if self.raw_selects == "*":
      self.vars_needed_for_select = { self.parse_var(var_name) for var_name in self.block.get_ordered_vars() }         
      
    for i, triple in enumerate(self.triples): # Pass 0: Simple union blocks
      if isinstance(triple, UnionBlock) and triple.simple_union_triples:
        self.triples[i:i+1] = triple.simple_union_triples
        
    remnant_triples = set(self.triples)
        
    for triple in list(self.triples): # Pass 1: Determine var type and prop type
      if isinstance(triple, (Bind, Filter, Block)): continue
      s, p, o = triple
      
      triple.local_table_type = triple.table_type
      if s.name == "VAR":
        var = self.parse_var(s)
        var.update_type("objs")
        if (p.name == "IRI") and (p.storid == rdfs_subpropertyof) and (o.name == "IRI"):
          parent_prop = self.translator.world._get_by_storid(o.storid)
          if   isinstance(parent_prop, ObjectPropertyClass): var.prop_type = "objs"
          elif isinstance(parent_prop, DataPropertyClass):   var.prop_type = "datas"
      if p.name == "VAR": self.parse_var(p).update_type("objs")
      if o.name == "VAR": self.parse_var(o).update_type(triple.local_table_type)
      
    for triple in list(self.triples): # Pass 2: Determine var type, which may be changed due to prop type
      if isinstance(triple, (Bind, Filter, Block)): continue
      s, p, o = triple
      if (triple.local_table_type == "quads") and (o.name == "VAR"): triple.local_table_type = self.parse_var(o).type
      if (triple.local_table_type == "quads") and (p.name == "VAR"): triple.local_table_type = self.parse_var(p).prop_type # Repeat (table.type == "quads") condition, since table.type may have been changed by the previous if block
      if o.name == "VAR": self.parse_var(o).update_type(triple.local_table_type)
      
    non_preliminary_triples = []
    for triple in list(self.triples): # Pass 3: Create recursive preliminary selects
      if isinstance(triple, Block): continue
      if isinstance(triple, (Bind, Filter)):
        non_preliminary_triples.append(triple)
        continue
      s, p, o = triple
      triple.consider_s = triple.consider_p = triple.consider_o = True
      
      if p.modifier:
        if   (s.name != "VAR"): fixed = "s"
        elif (o.name != "VAR"): fixed = "o"
        else:
          fix_levels = self.get_fix_levels([self.parse_var(s), self.parse_var(o)], triple)
          if fix_levels[self.parse_var(s)] >= fix_levels[self.parse_var(o)]: fixed = "s"
          else:                                                              fixed = "o"
          
        non_fixed = "o" if fixed == "s" else "s"
        vars = []
        if   (s.name == "VAR") and (fixed == "s"): fixed_var = s
        elif (o.name == "VAR") and (fixed == "o"): fixed_var = o
        else:                                      fixed_var = None
        if fixed_var: vars.append(self.parse_var(fixed_var))
        if  p.name == "VAR": vars.append(self.parse_var(p))
        
        prelim_triples = self.extract_triples(self.triples, vars, triple)
        
        remnant_triples.remove(triple)
        remnant_triples.difference_update(prelim_triples)
        prelim = self.translator.get_recursive_preliminary_select(triple, fixed, fixed_var, prelim_triples)
        triple.local_table_type = prelim.name
        triple.consider_p = False
        if not(fixed_var and prelim_triples):
          if fixed == "s": triple.consider_s = False
          else:            triple.consider_o = False
          
      else:
        non_preliminary_triples.append(triple)
        
    selected_non_preliminary_triples = frozenset(self.extract_triples(non_preliminary_triples, self.vars_needed_for_select))
    selected_non_preliminary_triples = selected_non_preliminary_triples | remnant_triples
    
    
    vars_needing_binding = set(self.vars_needed_for_select)
    for triple in self.triples: # Pass 4: compute number of bindings per variable
      if   isinstance(triple, (Bind, Filter)):
        for var_name in triple.var_names:
          vars_needing_binding.add(self.parse_var(var_name))
        continue
      if   isinstance(triple, Block): continue
      s, p, o = triple
      if (not p.modifier) and (not triple in selected_non_preliminary_triples): continue
      if triple.local_table_type.startswith("prelim") and not triple.consider_p: continue
      for (x, consider) in zip(triple, [triple.consider_s, triple.consider_p, triple.consider_o]):
        if consider and (x.name == "VAR"):
          x = self.parse_var(x)
          x.nb_table += 1
          
    for triple in self.triples: # Pass 5: pre-register bindings for preliminary table
      if   isinstance(triple, (Bind, Filter, Block)): continue
      s, p, o = triple
      if (not p.modifier) and (not triple in selected_non_preliminary_triples): continue
      
      triple.to_skip = False
      if self.raw_selects == "*": continue
      if triple.local_table_type.startswith("prelim") and not triple.consider_p:
        if (s.name == "VAR") and (self.parse_var(s) in vars_needing_binding) and self.parse_var(s).nb_table == 0: continue
        if (o.name == "VAR") and (self.parse_var(o) in vars_needing_binding) and self.parse_var(o).nb_table == 0: continue
        triple.to_skip = True
        extra = ""
        if p.modifier == "+": extra = " WHERE nb>0"
        if s.name == "VAR":
          s = self.parse_var(s)
          s.bindings.insert(0, """IN (SELECT s FROM %s%s)""" % (triple.local_table_type, extra))
        if o.name == "VAR":
          o = self.parse_var(o)
          o.bindings.insert(0, """IN (SELECT o FROM %s%s)""" % (triple.local_table_type, extra))
          
    for triple in self.triples: # Pass 6: create triples tables and conditions
      if   isinstance(triple, Bind):
        self.parse_bind(triple)
        continue
      elif isinstance(triple, Filter):
        self.parse_filter(triple)
        continue
      elif isinstance(triple, Block):
        if   isinstance(triple, SimpleTripleBlock) and (len(triple) == 0): continue # Empty
        elif isinstance(triple, FilterBlock):
          sub = self.translator.new_sql_query(None, triple, nested_inside = self)
          self.add_subquery(sub)
          continue
        else:
          sub = self.translator.new_sql_query(None, triple, preliminary = True, nested_inside = self, copy_vars = True)
          self.add_subquery(sub)
          continue
      if triple.to_skip: continue
      
      s, p, o = triple
      if (not p.modifier) and (not triple in selected_non_preliminary_triples): continue
      
      if self.name == "main": select_name = ""
      else:                   select_name = "%s_" % self.name
      table = Table("%sq%s" % (select_name, self.next_table_id), triple.local_table_type)
      if triple.optional:
        table.join = "LEFT JOIN"
        conditions = table.join_conditions
      else:
        conditions = self.conditions
      self.next_table_id += 1
      self.tables.append(table)
      self.name_2_table[table.name] = table
      
      if triple.consider_s: self.create_conditions(conditions, table, "s", s)
      if triple.consider_p: self.create_conditions(conditions, table, "p", p, triple.likelihood_p)
      if triple.consider_o: self.create_conditions(conditions, table, "o", o, triple.likelihood_o)
      
      if p.modifier == "+": conditions.append("%s.nb>0"  % table.name)

  def get_fix_levels(self, vars0, exclude_triple = None):
    vars0_names = { var.name for var in vars0 }
    fix_levels  = defaultdict(float)
    fix_triples = {}
    
    def fix(var, triple, via_vars, w = 1.0):
      nonlocal changed
      
      w0 = min((0.5 * fix_levels[via_var] for via_var in via_vars), default = 1.0)
      w  = w * w0
      
      if w > fix_levels[var]:
        changed          = True
        fix_levels [var] = w
        fix_triples[var] = l = { triple }
        for via_var in via_vars: l.update(fix_triples[via_var])
        
    def scan_triple(triples, w = 1):
      for triple in self.triples:
        if triple is exclude_triple: continue
        
        if   isinstance(triple, Triple):
          if   len(triple.var_names) == 1:
            var = self.parse_var(tuple(triple.var_names)[0])
            fix(var, triple, [], w)
            
          elif len(triple.var_names) == 2:
            if triple.var_names != vars0_names:
              vars = [self.parse_var(var_name) for var_name in triple.var_names]
              for v1, v2 in [(vars[0], vars[1]), (vars[1], vars[0])]:
                fix(v1, triple, [v2], w)
                
        elif isinstance(triple, Filter):
          pass
        elif isinstance(triple, Bind):
          var = self.parse_var(triple.var)
          fix(var, triple, [self.parse_var(var_name) for var_name in triple.var_names], w)
          
        elif isinstance(triple, UnionBlock):
          for alternative in triple: scan_triple(alternative, w / len(triple))
          
    while True:
      changed = False
      scan_triple(self.triples)
      if not changed: break
      for var in vars0:
        if not var in fix_levels: break
      else:
        break
      
    return fix_levels
      
  def extract_triples(self, triples, vars, except_triple = None):
    #print("   EXTRACT", triples, vars, except_triple)
    var_names = { var.name for var in vars }
    while True:
      r = [triple for triple in triples if (not triple == except_triple) and isinstance(triple, (Triple, Filter, Bind)) and (not triple.var_names.isdisjoint(var_names))]
      var_names2 = { var_name for triple in r for var_name in triple.var_names }
      if var_names2 == var_names:
        #print("   EXTRACTED", r)
        return r
      var_names = var_names2
      
  def create_conditions(self, conditions, table, n, x, likelihood = None):
    if isinstance(x, SpecialCondition):
      x.create_conditions(conditions, table, n)
    else:
      sql, sql_type, sql_d, sql_d_type = self._to_sql(x)
      
      if table.query: # If datatype is 0 (=auto), disable datatype conditions and replace it by IS NOT NULL
        for column in table.query.columns:
          if column.name == "%sd" % n[:-1]:
            if str(column.binding) == "0":
              sql_d = None
              conditions.append("%s.%sd IS NOT NULL" % (table.name, n[:-1])) # Datatype part
            break
          
      if not sql is None:
        if isinstance(sql, str) and sql.startswith("IN "): operator = " "; sqls = self.parse_var(x).bindings
        else:                                              operator = "="; sqls = [sql]
        
        for sql in sqls:
          condition = "%s.%s%s%s"  % (table.name, n, operator, sql)
          if likelihood is None: conditions.append(condition)
          else:                  conditions.append("LIKELIHOOD(%s, %s)" % (condition, likelihood))
      if not sql_d is None:
        conditions.append("%s.%sd=%s" % (table.name, n[:-1], sql_d)) # Datatype part
        
      if x.name == "VAR": x = self.vars[x.value]
      if   isinstance(x, Variable):
        if not x.initial_query: x.initial_query = self
        x.bindings.append("%s.%s" % (table.name, n))
        
  def is_fixed(self, x):
    if x.name != "VAR": return True
    x = self.parse_var(x)
    if x.initial_query and (x.initial_query is not self): return True
    if x.bindings: return True
    return False
  
  def finalize_columns(self):
    selected_parameter_index = 0
    i = j = 0
    if self.raw_selects == "*": selects = [self.vars[var] for var in self.block.get_ordered_vars() if not var.startswith("??")]      
    else:                       selects = self.raw_selects
    
    def do_select(select):
      nonlocal selected_parameter_index
      if isinstance(select, str) and select.startswith("?"): select = self.vars[select]
      sql, sql_type, sql_d, sql_d_type = self._to_sql(select)
      
      if   isinstance(select, rply.Token) and (select.name == "VAR"): var_name = select.value
      elif isinstance(select, Variable):                              var_name = select.name
      else:                                                           var_name = None
      
      if sql == "?":
        self.parameters.insert(selected_parameter_index, select.number)
        selected_parameter_index += 1

      return var_name, sql, sql_type, sql_d, sql_d_type
    
    for select in selects:
      i += 1
      if isinstance(select, SimpleUnion):
        self.select_simple_union = True
        sql = []
        for select_item in select.items:
          var_name, sql_i, sql_type, sql_d, sql_d_type = do_select(select_item)
          sql.append(sql_i)
        sql_d = None # SimpleUnion is only supported for object here
        
      else:
        var_name, sql, sql_type, sql_d, sql_d_type = do_select(select)
        
      if sql is None:
        if not self.name: # Inside a UNION => the variable is not available in this alternative of th UNION
          sql = "NULL"
          sql_type = "objs"
        else:
          raise ValueError("Cannot select '%s'!" % select)
      self.columns.append(Column(var_name, sql_type,   sql,   "col%s_o" % i, j)); j += 1
      if not sql_d is None: self.columns.append(Column(var_name, sql_d_type, sql_d, "col%s_d" % i, j)); j += 1
      
  def _to_sql(self, x):
    if isinstance(x, rply.Token) and (x.name == "VAR"): x = self.parse_var(x)

    if   isinstance(x, str): return x, "value", None, None
    elif isinstance(x, Variable):
      if not x.bindings: return None, None, None, None
      binding = x.get_binding(self)
      
      if   x.type == "objs":  return binding, "objs", None, None
      else:
        if not x.fixed_datatype is None:
          if isinstance(x.fixed_datatype, Variable):
            dropit, dropit, other_sql_d, other_sql_d_type = self._to_sql(x.fixed_datatype)
            return binding, "datas", other_sql_d, other_sql_d_type
          else:
            return binding, "datas", x.fixed_datatype, "datas"
        type = "datas" if x.type == "datas" else "quads"
        return binding, type, "%sd" % binding[:-1], type
    elif x.name == "IRI":   return x.storid, "objs", None, None
    elif x.name == "PARAM": return "?%s" % x.number, "objs", None, None # XXX data parameter
    else:
      if   x.name == "DATA":            return x.value, "value", x.datatype, "value"
      elif isinstance(x.value, locstr): return x.value, "value", "'@%s'" % x.value.lang, "value"
      else:                             return x.value, "value", None, None
      
    
class SQLCompoundQuery(object):
  recursive = False
  def __init__(self, name, parent):
    self.name                    = name
    self.parent                  = parent
    self.translator              = CURRENT_TRANSLATOR.get()
    self.queries                 = []
    self.preliminary             = False
    self.optional                = False
    
  def __repr__(self): return "<%s '%s'>" % (self.__class__.__name__, self.sql())
  
  def append(self, query, operator = ""):
    query.operator = operator
    self.queries.append(query)
    
  def sql(self):
    sql = ""
    for i, query in enumerate(self.queries):
      if i != 0: sql += "\n%s\n" % query.operator
      sql += query.sql()
    if self.preliminary: return """%s(%s) AS (%s)""" % (self.name, ", ".join(column.name for column in self.columns), sql)
    return sql
  
  def finalize_compounds(self):
    has_d = set()
    for query in self.queries:
      for column in query.columns:
        if column.name.endswith("d"):
          has_d.add(column.name.split("_", 1)[0])
          
    for query in self.queries:
      for i, column in enumerate(query.columns):
        if column.name.endswith("o") and (column.name.split("_", 1)[0] in has_d):
          column.type = "quads"
          if (column is query.columns[-1]) or (not query.columns[i+1].name.endswith("d")):
            query.columns.insert(i + 1, Column(column.var, "quads", "NULL", "%sd" % column.name[:-1], i + 1))
      for i, column in enumerate(query.columns): # Re-index columns
        column.index = i
    self.columns = self.queries[0].columns
    
    
class SQLNestedQuery(SQLQuery):
  def __init__(self, name):
    SQLQuery.__init__(self, name)
    self.exists = True
    
  def finalize_columns(self):
    self.columns = [Column(None, "datas", "1", "col1_o", 1)]
    
  def sql(self):
    #extra_sql = self.extra_sql
    #self.extra_sql = ""
    
    #sql = "(%s)" % SQLQuery.sql(self)
    
    #if extra_sql: sql += " %s" % extra_sql
    #self.extra_sql = extra_sql
    
    sql = SQLQuery.sql(self)

    if _DEPRIORIZE_SUBQUERIES_OPT:
      if   self.exists == True:  sql =     "one.i= EXISTS(%s)" % sql
      elif self.exists == False: sql = "NOT one.i= EXISTS(%s)" % sql
      else:                      sql = "(%s)" % sql
    else:
      if   self.exists == True:  sql =     "EXISTS(%s)" % sql
      elif self.exists == False: sql = "NOT EXISTS(%s)" % sql
      else:                      sql = "(%s)" % sql
      
    return sql
  
  def __str__(self): return self.sql()
  

class SQLRecursivePreliminaryQuery(SQLQuery):
  def __init__(self, name, triple, fixed, fixed_var):
    s, p, o = triple
    translator = CURRENT_TRANSLATOR.get()
    self.fixed        = fixed
    self.fixed_var    = fixed_var
    self.non_fixed    = "o" if fixed == "s" else "s"
    
    if isinstance(p, NegatedPropSetPath): self.need_d = True
    else:                                 self.need_d = (p.modifier == "?") and not isinstance(triple.Prop, ObjectPropertyClass)
    
    self.need_orig    = not self.fixed_var is None # XXX Optimizable
    self.need_nb      = p.modifier != "*"
    
    SQLQuery.__init__(self, "%s_%s" % (name, "quads" if self.need_d else "objs"))
    self.recursive    = True
    self.preliminary  = True
    
  def build(self, triple, prelim_triples):
    s, p, o = triple
    column_names = [self.non_fixed] + ["d"] * self.need_d + [self.fixed] * self.need_orig + ["nb"] * self.need_nb
    if self.fixed_var and prelim_triples: value = self.fixed_var
    else:                                 value = s if self.fixed == "s" else o
    self.parse_selects([value] + ["NULL"] * self.need_d + [value] * self.need_orig + ["0"] * self.need_nb)
    self.parse_triples(prelim_triples)
    self.finalize_columns()
    self.set_column_names(column_names)
    
    p_conditions = []
    self.create_conditions(p_conditions, Table("q", "quads" if self.need_d else "objs"), "p", p)
    self.extra_sql = """
UNION
SELECT q.%s%s%s%s FROM %s q, %s rec WHERE %s %sAND q.%s=rec.%s""" % (
  self.non_fixed,
  ", q.d"                 if self.need_d    else "",
  ", rec.%s" % self.fixed if self.need_orig else "",
  ", rec.nb+1"            if self.need_nb   else "",
  "quads"                 if self.need_d    else "objs",
  self.name, " AND ".join(p_conditions),
  "AND rec.nb=0 " if p.modifier == "?" else "",
  self.fixed, self.non_fixed)
  
  



# -*- coding: utf-8 -*-
# Owlready2
# Copyright (C) 2013-2019 Jean-Baptiste LAMY
# LIMICS (Laboratoire d'informatique médicale et d'ingénierie des connaissances en santé), UMR_S 1142
# University Paris 13, Sorbonne paris-Cité, Bobigny, France

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from owlready2.namespace       import *
from owlready2.entity          import *
from owlready2.prop            import *
from owlready2.class_construct import *


class AllDisjoint(object):
  def __init__(self, entities, ontology = None, bnode = None):
    #if not CURRENT_NAMESPACES[-1] is None: self.ontology = CURRENT_NAMESPACES[-1].ontology
    #else:                                  self.ontology = ontology or entities[0].namespace.ontology
    self.ontology = (CURRENT_NAMESPACES.get() and CURRENT_NAMESPACES.get()[-1]) or ontology or entities[0].namespace.ontology
    
    # For AllDisjoint, storid can be either:
    #   * a blank node, when at least 3 entities are involved
    #   * a triple, when only 2 entities, e.g. (entity1.storid, owl_disjointwith, entity2.storid)
    
    if   isinstance(entities, int):
      assert isinstance(bnode, int)
      self.storid      = bnode
      self._list_bnode = entities
    elif isinstance(entities, tuple):
      self.storid      = entities
    else:
      self.storid      = bnode
      self._list_bnode = None
      self.entities = self.Properties = self.individuals = CallbackList(entities, self, AllDisjoint._callback)
      
    if not LOADING: self._create_triples()
    
  def __getattr__(self, attr):
    if attr == "entities":
      if isinstance(self.storid, int):
        r = self.ontology._parse_list(self._list_bnode)
      else:
        r = [self.ontology.world._to_python(self.storid[0]), self.ontology.world._to_python(self.storid[2])]
      r = CallbackList(r, self, AllDisjoint._callback)
      setattr(self, attr,r)
      return r
    return super().__getattribute__(attr)
  
  def _callback(self, old):
    if self.ontology:
      self._destroy_triples()
      self._create_triples ()
      
  def _destroy_triples(self):
    if   isinstance(self.storid, int):
      self.ontology._del_obj_triple_spo(self.storid, None, None)
      self.ontology._del_list(self._list_bnode)
    elif isinstance(self.storid, tuple):
      self.ontology._del_obj_triple_spo(*self.storid)
      
  def destroy(self): self._destroy_triples()
  
  def _create_triples(self):
    for entity in self.entities:
      if isinstance(entity, Construct): entity._set_ontology(self.ontology)
      
    if len(self.entities) == 2:
      if   isinstance(self.entities[0], ThingClass):
        self.storid = (self.entities[0].storid, owl_disjointwith, self.entities[1].storid)
        self.ontology._add_obj_triple_spo(*self.storid)
        return
      elif isinstance(self.entities[0], PropertyClass):
        self.storid = (self.entities[0].storid, owl_propdisjointwith, self.entities[1].storid)
        self.ontology._add_obj_triple_spo(*self.storid)
        return
      # It seems that there is no 1-1 relation for individuals
      # => continue
      
    if len(self.entities) >= 2:
      if not isinstance(self.storid, int): self.storid      = self.ontology.world.new_blank_node()
      if not self._list_bnode:             self._list_bnode = self.ontology.world.new_blank_node()
      
      if   isinstance(self.entities[0], ThingClass):
        self.ontology._add_obj_triple_spo(self.storid, rdf_type, owl_alldisjointclasses)
        self.ontology._add_obj_triple_spo(self.storid, owl_members, self._list_bnode)
        
      elif isinstance(self.entities[0], PropertyClass):
        self.ontology._add_obj_triple_spo(self.storid, rdf_type, owl_alldisjointproperties)
        self.ontology._add_obj_triple_spo(self.storid, owl_members, self._list_bnode)
        
      else: # Individuals
        self.ontology._add_obj_triple_spo(self.storid, rdf_type, owl_alldifferent)
        self.ontology._add_obj_triple_spo(self.storid, owl_distinctmembers, self._list_bnode)
        
      self.ontology._set_list(self._list_bnode, self.entities)
      
  def __repr__(self):
    if self.ontology != self.entities[0].namespace.ontology: onto = ", ontology = %s" % self.ontology
    else:                                                    onto = ""
    return "AllDisjoint([%s]%s)" % (", ".join(repr(Class) for Class in self.entities), onto)
    
AllDifferent = AllDisjoint


def partition(mother, children):
  mother.is_a.append(Or(children))
  AllDisjoint(children)

  
# class DisjointUnion(LogicalClassConstruct):
#   _owl_op = owl_disjointunion
#   is_a    = ()
  
#   def _satisfied_by(self, x):
#     for Class in self.Classes:
#       if Class._satisfied_by(x): return True
#     return False
  
#   def __repr__(self):
#     s = []
#     for x in self.Classes:
#       if isinstance(x, LogicalClassConstruct): s.append("(%s)" % x)
#       else:                                    s.append(repr(x))
#     return "%s([%s])" % (self.__class__.__name__, ", ".join(s))



# -*- coding: utf-8 -*-
# Owlready2
# Copyright (C) 2019 Jean-Baptiste LAMY
# LIMICS (Laboratoire d'informatique médicale et d'ingénierie des connaissances en santé), UMR_S 1142
# University Paris 13, Sorbonne paris-Cité, Bobigny, France

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import sys, os, types, zipfile, gzip, io
from collections import defaultdict, Counter
from owlready2 import *

iso_3_letter_2_2_letter_lang = {'ABK': 'ab', 'AAR': 'aa', 'AFR': 'af', 'ALB': 'sq', 'SQI': 'sq', 'AMH': 'am', 'ARA': 'ar', 'ARG': 'an', 'ARM': 'hy', 'HYE': 'hy', 'ASM': 'as', 'AVE': 'ae', 'AYM': 'ay', 'AZE': 'az', 'BAK': 'ba', 'BAQ': 'eu', 'EUS': 'eu', 'BEL': 'be', 'BEN': 'bn', 'BIH': 'bh', 'BIS': 'bi', 'BOS': 'bs', 'BRE': 'br', 'BUL': 'bg', 'BUR': 'my', 'MYA': 'my', 'CAT': 'ca', 'CHA': 'ch', 'CHE': 'ce', 'CHI': 'zh', 'ZHO': 'zh', 'CHU': 'cu', 'CHV': 'cv', 'COR': 'kw', 'COS': 'co', 'SCR': 'hr', 'HRV': 'hr', 'CZE': 'cs', 'CES': 'cs', 'DAN': 'da', 'DIV': 'dv', 'DUT': 'nl', 'NLD': 'nl', 'DZO': 'dz', 'ENG': 'en', 'EPO': 'eo', 'EST': 'et', 'FAO': 'fo', 'FIJ': 'fj', 'FIN': 'fi', 'FRE': 'fr', 'FRA': 'fr', 'GLA': 'gd', 'GLG': 'gl', 'GEO': 'ka', 'KAT': 'ka', 'GER': 'de', 'DEU': 'de', 'GRE': 'el', 'ELL': 'el', 'GRN': 'gn', 'GUJ': 'gu', 'HAT': 'ht', 'HAU': 'ha', 'HEB': 'he', 'HER': 'hz', 'HIN': 'hi', 'HMO': 'ho', 'HUN': 'hu', 'ICE': 'is', 'ISL': 'is', 'IDO': 'io', 'IND': 'id', 'INA': 'ia', 'ILE': 'ie', 'IKU': 'iu', 'IPK': 'ik', 'GLE': 'ga', 'ITA': 'it', 'JPN': 'ja', 'JAV': 'jv', 'KAL': 'kl', 'KAN': 'kn', 'KAS': 'ks', 'KAZ': 'kk', 'KHM': 'km', 'KIK': 'ki', 'KIN': 'rw', 'KIR': 'ky', 'KOM': 'kv', 'KOR': 'ko', 'KUA': 'kj', 'KUR': 'ku', 'LAO': 'lo', 'LAT': 'la', 'LAV': 'lv', 'LIM': 'li', 'LIN': 'ln', 'LIT': 'lt', 'LTZ': 'lb', 'MAC': 'mk', 'MKD': 'mk', 'MLG': 'mg', 'MAY': 'ms', 'MSA': 'ms', 'MAL': 'ml', 'MLT': 'mt', 'GLV': 'gv', 'MAO': 'mi', 'MRI': 'mi', 'MAR': 'mr', 'MAH': 'mh', 'MOL': 'mo', 'MON': 'mn', 'NAU': 'na', 'NAV': 'nv', 'NDE': 'nd', 'NBL': 'nr', 'NDO': 'ng', 'NEP': 'ne', 'SME': 'se', 'NOR': 'no', 'NOB': 'nb', 'NNO': 'nn', 'NYA': 'ny', 'OCI': 'oc', 'ORI': 'or', 'ORM': 'om', 'OSS': 'os', 'PLI': 'pi', 'PAN': 'pa', 'PER': 'fa', 'FAS': 'fa', 'POL': 'pl', 'POR': 'pt', 'PUS': 'ps', 'QUE': 'qu', 'ROH': 'rm', 'RUM': 'ro', 'RON': 'ro', 'RUN': 'rn', 'RUS': 'ru', 'SMO': 'sm', 'SAG': 'sg', 'SAN': 'sa', 'SRD': 'sc', 'SCC': 'sr', 'SRP': 'sr', 'SNA': 'sn', 'III': 'ii', 'SND': 'sd', 'SIN': 'si', 'SLO': 'sk', 'SLK': 'sk', 'SLV': 'sl', 'SOM': 'so', 'SOT': 'st', 'SPA': 'es', 'SUN': 'su', 'SWA': 'sw', 'SSW': 'ss', 'SWE': 'sv', 'TGL': 'tl', 'TAH': 'ty', 'TGK': 'tg', 'TAM': 'ta', 'TAT': 'tt', 'TEL': 'te', 'THA': 'th', 'TIB': 'bo', 'BOD': 'bo', 'TIR': 'ti', 'TON': 'to', 'TSO': 'ts', 'TSN': 'tn', 'TUR': 'tr', 'TUK': 'tk', 'TWI': 'tw', 'UIG': 'ug', 'UKR': 'uk', 'URD': 'ur', 'UZB': 'uz', 'VIE': 'vi', 'VOL': 'vo', 'WLN': 'wa', 'WEL': 'cy', 'CYM': 'cy', 'FRY': 'fy', 'WOL': 'wo', 'XHO': 'xh', 'YID': 'yi', 'YOR': 'yo', 'ZHA': 'za', 'ZUL': 'zu'}

#ordered_srcs = ['SRC', 'SNOMEDCT_US', 'ICD10', 'ICPC', 'MDR', 'LNC', 'MSH', 'AIR', 'ALT', 'AOD', 'AOT', 'ATC', 'BI', 'CCC', 'CCPSS', 'CCS', 'CCS_10', 'CDT', 'CHV', 'COSTAR', 'CPM', 'CPT', 'CSP', 'CST', 'CVX', 'DDB', 'DRUGBANK', 'DSM-5', 'DXP', 'FMA', 'GO', 'GS', 'HCDT', 'HCPCS', 'HCPT', 'HGNC', 'HL7V2.5', 'HL7V3.0', 'HPO', 'ICD10AE', 'ICD10AM', 'ICD10AMAE', 'ICD10CM', 'ICD10PCS', 'ICD9CM', 'ICF', 'ICF-CY', 'ICNP', 'ICPC2EENG', 'ICPC2ICD10ENG', 'ICPC2P', 'JABL', 'LCH', 'LCH_NW', 'MCM', 'MED-RT', 'MEDCIN', 'MEDLINEPLUS', 'MMSL', 'MMX', 'MTH', 'MTHCMSFRF', 'MTHHH', 'MTHICD9', 'MTHICPC2EAE', 'MTHICPC2ICD10AE', 'MTHMST', 'MTHSPL', 'MVX', 'NANDA-I', 'NCBI', 'NCI', 'NCI_BRIDG', 'NCI_BioC', 'NCI_CDC', 'NCI_CDISC', 'NCI_CDISC-GLOSS', 'NCI_CRCH', 'NCI_CTCAE', 'NCI_CTCAE_3', 'NCI_CTCAE_5', 'NCI_CTEP-SDC', 'NCI_CTRP', 'NCI_CareLex', 'NCI_DCP', 'NCI_DICOM', 'NCI_DTP', 'NCI_FDA', 'NCI_GAIA', 'NCI_GENC', 'NCI_ICH', 'NCI_JAX', 'NCI_KEGG', 'NCI_NCI-GLOSS', 'NCI_NCI-HGNC', 'NCI_NCI-HL7', 'NCI_NCPDP', 'NCI_NICHD', 'NCI_PI-RADS', 'NCI_PID', 'NCI_RENI', 'NCI_UCUM', 'NCI_ZFin', 'NDDF', 'NDFRT', 'NDFRT_FDASPL', 'NDFRT_FMTSME', 'NEU', 'NIC', 'NOC', 'NUCCPT', 'OMIM', 'OMS', 'PCDS', 'PDQ', 'PNDS', 'PPAC', 'PSY', 'QMR', 'RAM', 'RCD', 'RCDAE', 'RCDSA', 'RCDSY', 'RXNORM', 'SNM', 'SNMI', 'SNOMEDCT_VET', 'SOP', 'SPN', 'ULT', 'UMD', 'USP', 'USPMG', 'UWDA', 'VANDF', 'WHO']
#terminology_2_priority = { terminology : i * 2 for i, terminology in enumerate(ordered_srcs) }


def create_model():
  PYM = get_ontology("http://PYM/")
  PYM.python_module = "owlready2.pymedtermino2.model"
  
  with PYM:
    class Concept(Thing): pass
    
    class originals(Concept >> Concept):
      class_property_type = ["some", "only"]
      
    class unifieds(Concept >> Concept):
      class_property_type = ["some", "only"]
      inverse = originals
      
    class SemanticType(Thing): pass
    
    class Group(Thing): pass
    
    class groups(Concept >> Group): pass
    
    class synonyms(AnnotationProperty): pass
    
    class terminology(AnnotationProperty): pass
    
    class definitions(AnnotationProperty): pass
    
  return PYM



normstr_storid = default_world._abbreviate("http://www.w3.org/2001/XMLSchema#normalizedString")
bool_storid    = default_world._abbreviate("http://www.w3.org/2001/XMLSchema#boolean")


def parse_mrrank(PYM, terminologies, langs, importer, f, remnant = ""):
  for line in f:
    if remnant: line = "%s%s" % (remnant, line); remnant = ""
    try:
      rank, terminology, tty, suppress, _dropit = line.split("|")
    except: return line
    
    if (importer.terminologies and (not terminology in importer.terminologies) and (terminology != "SRC")): continue
    if suppress in importer.remove_suppressed: continue
    
    importer.tty_2_priority[terminology, tty] = int(rank)
  
    
def parse_mrconso(PYM, terminologies, langs, importer, f, remnant = ""):
  for line in f:
    if remnant: line = "%s%s" % (remnant, line); remnant = ""
    try:
      cui, lang, ts, lui, stt, sui, is_pref, aui, saui, scui, \
        sdui, terminology, tty, orig_code, term, srl, suppress, cvf, _dropit = line.split("|")
    except: return line
    
    lang = iso_3_letter_2_2_letter_lang[lang]
    
    if langs and (not lang in importer.langs): continue
    
    if (terminology == "SRC") and orig_code.startswith("V-"): orig_code = orig_code[2:]
    if (importer.terminologies and (not terminology in importer.terminologies)):
      if not ((terminology == "SRC") and ((orig_code == "SRC") or (orig_code in importer.terminologies))):
        continue
    
    if   orig_code == "NOCODE":
      orig_code = importer.next_arbitrary_code
      importer.next_arbitrary_code += 1
    elif orig_code == "R40-F46.9": orig_code = "R40-R46.9" # Error in UMLS!
    elif orig_code == "R90-F94.9": orig_code = "R90-R94.9" # Error in UMLS!
    
    if importer.extract_cui:
      cui               = importer._abbreviate("http://PYM/CUI/%s" % cui)
    orig                = importer._abbreviate("http://PYM/%s/%s" % (terminology, orig_code))
    
    importer.aui_2_orig[aui] = orig
    
    if suppress in importer.remove_suppressed: continue
    
    if not terminology in importer.created_terminologies:
      importer.created_terminologies.add(terminology)
      importer.terminology_2_parents[terminology] = defaultdict(set)
      
    if importer.extract_cui:
      if not cui in importer.cui_2_origs:
        importer.objs.append((cui, rdf_type, owl_class))
        #importer.objs.append((cui, rdfs_subclassof, PYM.UnifiedConcept.storid))
        importer.objs.append((cui, rdfs_subclassof, importer.CUI))
        importer.objs.append((cui, PYM.terminology.storid, importer.CUI))
        importer.cui_2_origs   [cui] = { orig }
        importer.cui_2_terms   [cui] = []
      else:
        importer.cui_2_origs[cui].add(orig)
        
    if not orig in importer.orig_2_terminology:
      terminology_storid = importer._abbreviate("http://PYM/SRC/%s" % terminology)
      importer.objs.append((orig, rdf_type, owl_class))
      importer.objs.append((orig, PYM.terminology.storid, terminology_storid))
      importer.orig_2_terminology[orig] = terminology      
      importer.orig_2_terms      [orig] = []
      if importer.extract_cui: importer.orig_2_cuis[orig] = set()
      
      if (terminology == "SRC") and (orig_code != "SRC"):
        importer.terminology_2_parents["SRC"][orig].add(importer._abbreviate("http://PYM/SRC/SRC"))
        
    label_priority = importer.tty_2_priority[terminology, tty]
    importer.orig_2_terms[orig].append((label_priority, term, lang))
    
    if importer.extract_cui:
      importer.cui_2_terms[cui].append((label_priority, term, lang))
      if not cui in importer.orig_2_cuis[orig]:
        importer.orig_2_cuis[orig].add(cui)
        importer.restrict(orig, SOME, PYM.unifieds.storid,  cui)
        importer.restrict(cui,  SOME, PYM.originals.storid, orig)
        
    importer.check_insert()

  
# def parse_mrhier(PYM, terminologies, langs, importer, f, remnant = ""):
#   for line in f:
#     if remnant: line = "%s%s" % (remnant, line); remnant = ""
#     try:
#       cui1, aui1, cxn, parent_aui, terminology, rela, hier, hcd, cvf, _dropit = line.split("|")
#     except: return line
    
#     parent_aui = hier.rsplit(".", 1)[-1]
    
#     child_orig  = importer.aui_2_orig.get(aui1)
#     if not child_orig: continue
#     parent_orig = importer.aui_2_orig.get(parent_aui)
#     if not parent_orig: continue
    
#     if (child_orig != parent_orig) and (child_orig in importer.orig_2_terminology):
#       importer.terminology_2_parents[importer.orig_2_terminology[child_orig]] [child_orig].add(parent_orig)
      
#     importer.check_insert()
    

def parse_mrrel(PYM, terminologies, langs, importer, f, remnant_previous = ""):
  if remnant_previous == "":
    remnant = ""
    previous = [None] * 17
  else:
    remnant, previous = remnant_previous
    
  for line in f:
    if remnant: line = "%s%s" % (remnant, line); remnant = ""
    current = line.split("|")
    if len(current) < 17: return line, previous
    
    current = [current[i] or previous[i] for i in range(len(previous))]
    cui1, aui1, stype1, rel, cui2, aui2, stype2, rela, rui, srui, \
      terminology, sl, group_i, direct, suppress, cvf, _dropit = current
    
    if suppress in importer.remove_suppressed:
      previous = current
      continue
    
    cui1 = cui1 or previous[0]
    direct = direct == "Y"
    
    if (rel == "PAR") or (rel == "CHD"):
      if rel == "PAR": cui1, aui1, stype1 = cui2, aui2, stype2
      
      if aui1 and aui2:
        orig1 = importer.aui_2_orig.get(aui1)
        orig2 = importer.aui_2_orig.get(aui2)
        if orig1 and (orig1 != orig2) and (orig2 in importer.orig_2_terminology):
          importer.terminology_2_parents[importer.orig_2_terminology[orig2]] [orig2].add(orig1)
          
    elif importer.extract_relations:
      orig1 = importer.aui_2_orig.get(aui1)
      orig2 = importer.aui_2_orig.get(aui2)
      if orig1 and orig2:
        if (not orig1 in importer.orig_2_terminology) or (not orig2 in importer.orig_2_terminology): continue
        prop = importer.get_prop((rela or rel), owl_object_property)
        importer.mkrel(orig2, prop, orig1, group_i, direct)
        
    importer.check_insert()
    previous = current
    
    
def parse_mrsat(PYM, terminologies, langs, importer, f, remnant_previous = ""):
  if remnant_previous == "":
    remnant = ""
    previous = [None] * 14
  else:
    remnant, previous = remnant_previous
    
  for line in f:
    if remnant: line = "%s%s" % (remnant, line); remnant = ""
    current = line.split("|")
    if len(current) < 14: return line, previous
    
    if not current[0]: current = [current[i] or previous[i] for i in range(len(previous))]
    cui, lui, sui, metaui, stype, code, atui, satui, atn, terminology, atv, suppress, cvf, _dropit = current
    
    if suppress in importer.remove_suppressed:
      previous = current
      continue
    if importer.terminologies and (not terminology in importer.terminologies):
      previous = current
      continue
    
    prop = importer.get_prop(atn.lower(), owl_annotation_property)
    
    if   metaui == "":
      if importer.extract_cui:
        cui = importer._abbreviate("http://PYM/CUI/%s" % cui)
        importer.datas.append((cui, prop, atv, 0))
        
    elif metaui.startswith("A"):
      orig = importer.aui_2_orig.get(metaui)
      if not orig is None:
        importer.datas.append((orig, prop, atv, 0))
        
    importer.check_insert()
    previous = current
    
    
def parse_mrdef(PYM, terminologies, langs, importer, f, remnant = ""):
  for line in f:
    if remnant: line = "%s%s" % (remnant, line); remnant = ""
    try:
      cui, aui, atui, satui, terminology, defin, suppress, cvf, _dropit = line.split("|")
    except: return line
    
    if (importer.terminologies and (not terminology in importer.terminologies)): continue
    if suppress in importer.remove_suppressed: continue

    orig = importer.aui_2_orig.get(aui)
    if not orig is None:
      importer.datas.append((orig, PYM.definitions.storid, defin, 0))
      
    importer.check_insert()


def parse_mrsty(PYM, terminologies, langs, importer, f, remnant = ""):
  for line in f:
    if remnant: line = "%s%s" % (remnant, line); remnant = ""
    try:
      cui, tui, stn, sty, atui, cvf, _dropit = line.split("|")
    except: return line
    
    sem = importer.semantic_types.get(tui)
    if sem is None:
      sem = importer._abbreviate("http://PYM/TUI/%s" % tui)
      importer.semantic_types[tui] = sem
      importer.datas.append((sem, label.storid, sty, 0))
      
    importer.objs.append((cui, rdfs_subclassof, sem))
    
    importer.check_insert()
    
def parse_srdef(PYM, terminologies, langs, importer, f, remnant = ""):
  for line in f:
    if remnant: line = "%s%s" % (remnant, line); remnant = ""
    try:
      rt, tui, term, stn, defin, ex, un, nh, abr, inv, _dropit = line.split("|")
    except: return line
    
    sem = importer._abbreviate("http://PYM/TUI/%s" % tui)
    importer.semantic_types[tui] = sem
    importer.datas.append((sem, label.storid, term, "@en"))
    importer.datas.append((sem, PYM.synonyms.storid, abr, "@en"))
    importer.datas.append((sem, PYM.definitions.storid, defin, 0))
    
    if inv:
      importer.objs.append((sem, owl_inverse_property, ))
      
    importer.check_insert()


    
def break_cycles(parents, terminology = ""):
  equivalences = set()
  
  nb = 0
  already = set()
  for node in parents:
    if node in already: continue
    nb += 1
    
    paths = [[node]]
    explored_paths = ""
    while paths:
      new_paths = []
      for path in paths:
        already.add(path[-1])
        next_nodes = parents.get(path[-1], [])
        for next_node in next_nodes:
          if next_node in path: # Cycle found
            cycle = frozenset(path[path.index(next_node):] + [next_node])
            equivalences.add(cycle)
          else:
            new_paths.append(path + [next_node])
      paths = new_paths

  equivalences2 = []
  for cycle in equivalences:
    for other in equivalences:
      if other == cycle: continue
      if other.issuperset(cycle): break
    else: # No superset
      equivalences2.append(cycle)
      
  
  del equivalences
  nb_cycles = len(equivalences2)
  print("   ", terminology, ":", nb_cycles, "cycles found:", ", ".join(str(set(i)) for i in equivalences2))
  
  new_parents = { k : set(v) for k, v in parents.items() }
  equivalences3 = []
  for cycle in equivalences2:
    d = Counter()
    if isinstance(next(iter(parents.values())), Counter):
      for i in cycle:
        for p in parents[i]:
          if p in cycle: d[p] += parents[i][p]
    else:
      for i in cycle:
        for p in parents[i]:
          if p in cycle: d[p] += 1
    mosts = d.most_common()
    if mosts[0][1] >= 2 * mosts[1][1]:
      print("        Breaking cycle", mosts, "by considering", mosts[0][0], "as parent of the others")
      for i in cycle:
        for j in cycle:
          if j in new_parents[i]: new_parents[i].remove(j)
      for i in cycle:
        if i == mosts[0][0]: continue
        new_parents[i].add(mosts[0][0])
        
    else:
      print("        Breaking cycle", mosts, "by considering all concepts as equivalent")
      for i in cycle:
        for j in cycle:
          if j in new_parents[i]: new_parents[i].remove(j)
      cycle_parents = set()
      for i in cycle: cycle_parents.update(new_parents[i])
      for i in cycle: new_parents[i].update(cycle_parents)
      
      equivalences3.append(cycle)
      
  return new_parents, equivalences3, nb_cycles


def finalize(PYM, importer):
  # print("Breaking CUI cycles...")
  # #open("/tmp/parents.txt", "w").write("parents = %s" % importer.cui_parents)
  # parents, equivalences = break_cycles(importer.cui_parents)
  # print("   ", len(equivalences), "cycles found")
  
  # for cui in importer.cui_parents:
  #   cui_parents = parents.get(cui)
  #   if cui_parents:
  #     for parent in cui_parents:
  #       importer.objs.append((cui, rdfs_subclassof, parent))
  #   else:
  #     importer.objs.append((cui, rdfs_subclassof, PYM.UnifiedConcept.storid))
  # importer.check_insert()
  
  # for cycle in equivalences:
  #   cycle = list(cycle)
  #   for other in cycle[1:]:
  #     importer.objs.append((cycle[0], owl_equivalentclass, other))
  # importer.check_insert()
  
  
  print("Breaking ORIG cycles...")
  terminology_2_origs = { terminology : [] for terminology in importer.terminology_2_parents }
  for orig, terminology in importer.orig_2_terminology.items():
    terminology_2_origs[terminology].append(orig)
    
  for terminology, parents in importer.terminology_2_parents.items():
    #open("/tmp/parents2.txt", "w").write("parents = %s" % parents)
    parents, equivalences, nb_cycles = break_cycles(parents, terminology)
    for orig in terminology_2_origs[terminology]:
      orig_parents = parents.get(orig)
      if orig_parents:
        for parent in orig_parents:
          importer.objs.append((orig, rdfs_subclassof, parent))
      else:
        terminology = importer.orig_2_terminology[orig]
        if terminology == "SRC":
          importer.objs.append((orig, rdfs_subclassof, PYM.Concept.storid))
        else:
          importer.objs.append((orig, rdfs_subclassof, importer._abbreviate("http://PYM/SRC/%s" % terminology)))
          
    importer.check_insert()

    for cycle in equivalences:
      cycle = list(cycle)
      for other in cycle[1:]:
        importer.objs.append((cycle[0], owl_equivalentclass, other))
    importer.check_insert()
  del importer.terminology_2_parents
  terminology_2_origs = None

  
  print("Finalizing only properties and restrictions...")
  for (x, prop), ys in importer.onlys.items():
    if   len(ys) == 1:
      importer.restrict(x, ONLY, prop, ys[0])
    elif ys:
      l  = importer.create_rdf_obj_list(ys)
      bn = importer.new_blank_node()
      importer.objs.append((bn, rdf_type, owl_class))
      importer.objs.append((bn, owl_unionof, l))
      importer.restrict(x, ONLY, prop, bn)
  del importer.onlys
  
  for prop in set(importer.direct_props.keys()) | set(importer.indirect_props.keys()):
    if importer.indirect_props[prop] > importer.direct_props[prop]:
      importer.datas.append((prop, owlready_class_property_type, "only", 0))
  del importer.direct_props
  del importer.indirect_props
      
      
  print("Finalizing CUI - ORIG mapping...")
  if importer.extract_cui:
    for cui, cui_origs in importer.cui_2_origs.items():
      cui_origs = list(cui_origs)
    
      if   len(cui_origs) == 1:
        importer.restrict(cui, ONLY, PYM.originals.storid, cui_origs[0])
        
      elif len(cui_origs)  > 0:
        l  = importer.create_rdf_obj_list(cui_origs)
        bn = importer.new_blank_node()
        importer.objs.append((bn, rdf_type, owl_class))
        importer.objs.append((bn, owl_unionof, l))
        importer.restrict(cui, ONLY, PYM.originals.storid, bn)
        
      importer.check_insert()
    del importer.cui_2_origs
    
    for orig, orig_cuis in importer.orig_2_cuis.items():
      orig_cuis = list(orig_cuis)
      
      if   len(orig_cuis) == 1:
        importer.restrict(orig, ONLY, PYM.unifieds.storid, orig_cuis[0])
        
      elif len(orig_cuis)  > 0:
        l  = importer.create_rdf_obj_list(orig_cuis)
        bn = importer.new_blank_node()
        importer.objs.append((bn, rdf_type, owl_class))
        importer.objs.append((bn, owl_unionof, l))
        importer.restrict(orig, ONLY, PYM.unifieds.storid, bn)
        
      importer.check_insert()
    del importer.orig_2_cuis
    
class _Importer(object):
  def __init__(self, PYM, terminologies, langs, extract_groups, extract_attributes, extract_relations, extract_definitions, remove_suppressed):
    self.PYM          = PYM
    self.terminologies = terminologies
    self.langs         = langs
    self.extract_groups      = extract_groups
    self.extract_attributes  = extract_attributes
    self.extract_relations   = extract_relations
    self.extract_definitions = extract_definitions
    self.remove_suppressed   = remove_suppressed
    
    self.tty_2_priority = defaultdict(int)
    self.created_terminologies = set()
    self.cui_2_origs = {}
    self.orig_2_cuis = {}
    self.orig_2_terminology = {}
    self.aui_2_orig  = {}
    self.props = set()
    self.terminology_2_parents = {}
    self.cui_parents = {}
    self.next_arbitrary_code = 1
    
    
    self.orig_2_terms = {}
    self.cui_2_terms  = {}
    self.partial_relations = {}
    self.relations = set()
    self.semantic_types = {}
    self.groups = {}
    self.extract_cui = (not terminologies) or ("CUI" in terminologies)
    self.onlys = defaultdict(list)
    self.indirect_props = Counter()
    self.direct_props   = Counter()
    
    self.objs, self.datas, self.on_prepare_obj, self.on_prepare_data, self.insert_objs, self.insert_datas, self.new_blank_node, self._abbreviate, self.on_finish = PYM.graph.create_parse_func(delete_existing_triples = False)
    
    if self.extract_cui:
      self.CUI = self._abbreviate("http://PYM/SRC/CUI")
      self.objs .append((self.CUI, rdfs_subclassof, PYM.Concept.storid))
      self.objs .append((self.CUI, PYM.terminology.storid, self._abbreviate("http://PYM/SRC/SRC")))
      self.datas.append((self.CUI, label.storid, "UMLS unified concepts (CUI)", 0))
      
      
  def after(self, parser): # Free some memory
    self.check_insert()
    if   parser == "MRCONSO":
      for orig, terms in self.orig_2_terms.items():
        terms.sort()
        self.datas.append((orig, label.storid, terms[-1][1], "@%s" % terms[-1][2]))
        for priority, term, lang in terms[:-1]:
          self.datas.append((orig, self.PYM.synonyms.storid, term, "@%s" % lang))
        self.check_insert()
      del self.orig_2_terms
      
      for cui, terms in self.cui_2_terms.items():
        terms.sort()
        self.datas.append((cui, label.storid, terms[-1][1], "@%s" % terms[-1][2]))
        for priority, term, lang in terms[:-1]:
          self.datas.append((cui, self.PYM.synonyms.storid, term, "@%s" % lang))
        self.check_insert()
        
      #for cui, origs in self.cui_2_origs.items():
      #  terms = []
      #  for orig in origs:
      #    terms.extend(self.orig_2_terms[orig])
      #  terms.sort()
      #  self.datas.append((cui, label.storid, terms[-1][1], "@%s" % terms[-1][2]))
      #  for priority, term, lang in terms[:-1]:
      #    self.datas.append((cui, self.PYM.synonyms.storid, term, "@%s" % lang))
      #  self.check_insert()
      
      del self.cui_2_terms
      
    elif parser == "MRREL":
      del self.relations
      del self.partial_relations
      del self.groups
      
    elif parser == "MRSAT":
      del self.aui_2_orig
      
  def get_prop(self, name, type):
    prop = self._abbreviate("http://PYM/%s" % name)
    if not prop in self.props:
      self.objs.append((prop, rdf_type, type))
      self.props.add(prop)
    return prop
  
  def get_group(self, orig, i):
    group = self.groups.get((orig, i))
    if group is None:
      group = self.groups[orig, i] = self._abbreviate("http://PYM/groups/%s_%s" % (orig, i))
      self.objs.append((group, rdf_type, owl_class))
      self.objs.append((group, rdfs_subclassof, self.PYM.Group.storid))
      self.restrict(orig, SOME, self.PYM.groups.storid, group)
    return group
  
  def mkrel(self, orig2, prop, orig1, group_i, direct):
    if direct:
      self.direct_props[prop] += 1
      if not (orig2, prop, orig1) in self.relations: # Else, already done by a relation in another group
        self.restrict(orig2, SOME, prop, orig1)
        self.relations.add((orig2, prop, orig1))
      if group_i and self.extract_groups:
        self.restrict(self.get_group(orig2, group_i), SOME, prop, orig1)
    else:
      self.indirect_props[prop] += 1
      if not (orig2, prop, orig1) in self.relations: # Else, already done by a relation in another group
        self.onlys[orig2, prop].append(orig1)
        
  def restrict(self, a, qual, prop, b):
    bn = self.new_blank_node()
    self.objs.append((a, rdfs_subclassof, bn))
    self.objs.append((bn, rdf_type, owl_restriction))
    self.objs.append((bn, owl_onproperty, prop))
    self.objs.append((bn, qual, b))
    return bn
  
  def check_insert(self):
    if len(self.objs ) > 300000: self.insert_objs ()
    if len(self.datas) > 300000: self.insert_datas()
    
  def force_insert(self):
    self.insert_objs ()
    self.insert_datas()
    
  def create_rdf_obj_list(self, l):
    bnode = bnode0 = self.new_blank_node()
    for i in range(len(l)):
      self.objs.append((bnode, rdf_first, l[i]))
      if i < len(l) - 1:
        bnode_next = self.new_blank_node()
        self.objs.append((bnode, rdf_rest, bnode_next))
        bnode = bnode_next
      else:
        self.objs.append((bnode, rdf_rest, rdf_nil))
    return bnode0
    
    
def import_umls(umls_zip_filename, terminologies = None, langs = None, fts_index = True, extract_groups = True, extract_attributes = True, extract_relations = True, extract_definitions = True, remove_suppressed = "OEY"):
  if terminologies:
    terminologies = set(terminologies)
  if langs:
    langs = set(langs)

  PYM = create_model()
  default_world.save()
  
  #default_world.graph.set_indexed(False)
  
  importer = _Importer(PYM, terminologies, langs, extract_groups, extract_attributes, extract_relations, extract_definitions, remove_suppressed)
  
  parsers = [
    ("MRRANK",  parse_mrrank),
    ("MRCONSO", parse_mrconso),
  ]
  if extract_definitions: parsers.append(("MRDEF"  , parse_mrdef))
  parsers.append(("MRREL"  , parse_mrrel))
  if extract_attributes: parsers.append(("MRSAT"  , parse_mrsat))
  #if (not terminology) or ("CUI" in terminologies): parsers["MRSTY"] = parse_mrsty
  
  remnants = defaultdict(str)
  
  remnant_mrconso = remnant_mrrel = ""
  
  previous_parser = None
  
  if os.path.isdir(umls_zip_filename):
    print("Importing UMLS from directory %s with Python version %s.%s..." % (umls_zip_filename, sys.version_info.major, sys.version_info.minor))
    inner_filenames = sorted(os.listdir(umls_zip_filename))
    print("Files found in this directory: %s" % ", ".join(inner_filenames))
    if not "MRCONSO.RRF" in inner_filenames:
      print("WARNING: file MRCONSO.RRF not found!")
    for table_name, parser in parsers:
      for inner_filename in inner_filenames:
        if ("%s.RRF" % table_name) in inner_filename:
          if previous_parser != table_name: importer.after(previous_parser)
          print("  Parsing %s as %s..." % (inner_filename, table_name))
          remnants[table_name] = parser(PYM, terminologies, langs, importer,
                                        open(os.path.join(umls_zip_filename, inner_filename)),
                                        remnants[table_name])
          importer.force_insert()
          default_world.save()
          previous_parser = table_name
  else:
    print("Importing UMLS from Zip file %s with Python version %s.%s..." % (umls_zip_filename, sys.version_info.major, sys.version_info.minor))

    def parse_umls_zip(umls_inner_zip):
      nonlocal previous_parser
      inner_filenames = sorted(umls_inner_zip.namelist())
      for table_name, parser in parsers:
        for inner_filename in inner_filenames:
          if ("/%s.RRF" % table_name) in inner_filename:
            if previous_parser != table_name: importer.after(previous_parser)
            print("  Parsing %s as %s" % (inner_filename, table_name), end = "")
            if inner_filename.lower().endswith(".gz"):
              f = gzip.open(umls_inner_zip.open(inner_filename), "rt", encoding = "UTF-8")
              print(" with encoding %s" % f.encoding)
            else:
              f = io.TextIOWrapper(umls_inner_zip.open(inner_filename), "UTF-8")
              print()
            remnants[table_name] = parser(PYM, terminologies, langs, importer,
                                          f,
                                          remnants[table_name])
            importer.force_insert()
            default_world.save()
            previous_parser = table_name
            
    
    with zipfile.ZipFile(umls_zip_filename, "r") as umls_zip:
      filenames = sorted(umls_zip.namelist())
      
      #for filename in filenames:
      #  if filename.endwiths("/META/") or filename.endwiths("/META"):
      #    full_umls = False
      #    break
      #else: full_umls = True

      #if full_umls:
      parse_umls_zip(umls_zip)
      for filename in filenames:
        if filename.endswith("-meta.nlm"):
          print("Full UMLS release - importing UMLS from inner Zip file %s..." % filename)
          fp = umls_zip.open(filename)
          if not fp.seekable():
            raise Exception("PyMedTermino2 import require Python version >= 3.7 (but after import, the quadstore can be used by Python 3.6)")
          with zipfile.ZipFile(fp, "r") as umls_inner_zip:
            parse_umls_zip(umls_inner_zip)
            
      #else:
      #  parse_inner_zip(umls_zip)
        
      #for filename in filenames:
      #  if filename.endswith("-meta.nlm"):
      #    print("Importing UMLS from Zip file %s with Python version %s..." % (filename, sys.version))
      #    
      #    fp = umls_zip.open(filename)
      #    if not fp.seekable():
      #      raise Exception("PyMedTermino2 import require Python version >= 3.7 (but after import, the quadstore can be used by Python 3.6)")
      #    with zipfile.ZipFile(fp, "r") as umls_inner_zip:
      #      parse_inner_zip(umls_inner_zip)
            #inner_filenames = sorted(umls_inner_zip.namelist())
            #for table_name, parser in parsers:
            #  for inner_filename in inner_filenames:
            #    if ("/%s.RRF" % table_name) in inner_filename:
            #      if previous_parser != table_name: importer.after(previous_parser)
            #      print("  Parsing %s as %s" % (inner_filename, table_name), end = "")
            #      f = gzip.open(umls_inner_zip.open(inner_filename), "rt", encoding = "UTF-8")
            #      print(" with encoding %s" % f.encoding)
            #      remnants[table_name] = parser(PYM, terminologies, langs, importer,
            #                                    f,
            #                                    remnants[table_name])
            #      importer.force_insert()
            #      default_world.save()
            #      previous_parser = table_name
                  
  finalize(PYM, importer)
  importer.force_insert()
  importer.on_finish()
  importer = None # Free memory
  #default_world.save()
  
  #print("Indexing...")
  #default_world.graph.set_indexed(True)
  PYM = get_ontology("http://PYM/").load()
  
  if fts_index:
    print("FTS Indexing...")
    default_world.full_text_search_properties.append(label)
    default_world.full_text_search_properties.append(PYM.synonyms)
    
  import owlready2.pymedtermino2.model
  default_world.save()
  return PYM
